# MIT License

# Copyright (c) 2025 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Simulator:
    """The base class for the causal chamber simulators.

    Inheriting classes redefine the class attributes `inputs_names`
    and `outputs_names`, and the functions `__init__` and `_simulate`.

    """

    inputs_names = []
    outputs_names = []

    def __init__(self):
        """
        Initializes the Simulator instance. Re-defined by inheriting classes.
        """
        pass

    def simulate_from_inputs(self, df):
        """Runs the simulation using inputs from a given DataFrame. Passing a
        dataframe that doesn't define all inputs as columns will raise
        an error. Additional columns are ignored.

        Parameters
        ----------
        df : pandas.DataFrameA
            A pandas DataFrame containing the inputs in columns named
            as in `self.inputs_names`.

        Returns
        -------
            The result of the `_simulate` method, which should be
            implemented by subclasses.

        """
        # Take inputs from dataframe
        inputs = dict((k, v.values) for k, v in dict(df[self.inputs_names]).items())
        return self._simulate(**inputs, **self.parameters())

    def parameters(self):
        """
        Returns a dictionary with the simulator's parameters and their values.
        """
        return {}

    def _simulate(self):
        """
        A placeholder method for the simulation logic, to be implemented in subclasses.

        Raises
        ------
        NotImplementedError
            If not implemented in a subclass.
        """
        raise NotImplementedError()



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Functions for general sampling.
"""

import numpy as np


def random_walk(n, step_size=1, step_resolution=1, random_state=None):
    rng = np.random.default_rng(random_state)
    possible_transitions = np.arange(
        -step_size, step_size + step_resolution, step_resolution
    )
    steps = rng.choice(possible_transitions, size=n, replace=True)
    return np.cumsum(steps)


def random_walk_mirrored(
    n, minn, maxx, step_size=1, step_resolution=1, random_state=None
):
    rng = np.random.default_rng(random_state - 1)
    init = rng.choice(np.arange(minn, maxx + step_resolution, step_resolution))
    walk = init + random_walk(
        n,
        step_size=step_size,
        step_resolution=step_resolution,
        random_state=random_state,
    )
    # As long as walk is out of bounds
    while (
        maxx is not None
        and (walk > maxx).any()
        or minn is not None
        and (walk < minn).any()
    ):
        walk = np.minimum(walk, maxx) - np.maximum(walk, maxx) + maxx
        walk = np.maximum(walk, minn) - np.minimum(walk, minn) + minn
    return walk


# Deprecatedx
def discrete_rw(n, minn, maxx, step_sizes=[-1, 0, 1], p=None, seed=42):
    rng = np.random.default_rng(seed)
    walk = np.zeros(n)
    walk[0] = rng.integers(minn, maxx + 1)
    steps = rng.choice(step_sizes, p=p, size=n - 1, replace=True)
    for i, s in enumerate(steps):
        walk[i + 1] = min(max(walk[i] + s, minn), maxx)
    return walk


def smooth_discrete_rw(n, minn, maxx, step_sizes=[-1, 0, 1], p=None, lags=0, seed=42):
    walk = discrete_rw(n + lags, minn, maxx, step_sizes, p, seed)
    walk = pd.DataFrame(walk).rolling(lags).mean()[lags:]
    return walk.values



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np


# --------------------------------------------------------------------
# Functions to draw the hexagon


def hexagon_vertices(center_x, center_y, radius, offset):
    vertices = []
    for angle in np.arange(0, 2 * np.pi, np.pi / 3):
        x = center_x + radius * np.cos(angle + offset)
        y = center_y + radius * np.sin(angle + offset)
        vertices.append([x, y])
    return np.array(vertices)


def coord_grid(image_size):
    X = np.tile(np.arange(image_size), (image_size, 1))
    Y = X.T[::-1, :]
    grid = np.array([X, Y])
    return np.transpose(grid, (1, 2, 0))


def hexagon_mask(center_x, center_y, radius, offset, image_size):
    mask = np.zeros((image_size, image_size))
    image_points = coord_grid(image_size)
    vertices = hexagon_vertices(center_x, center_y, radius, offset) * image_size
    # Compute cross products for a segment and all points
    cross_prods = []
    for i, vertex in enumerate(vertices):
        segment = vertex - vertices[(i + 1) % len(vertices)]
        vertex_to_points = image_points - vertex
        cross = np.cross(vertex_to_points, segment)
        cross_prods.append(cross)
    cross_prods = np.array(cross_prods)
    all_neg = (cross_prods <= 0).all(axis=0)
    all_pos = (cross_prods > 0).all(axis=0)
    mask = np.logical_or(all_neg, all_pos)
    return mask


# --------------------------------------------------------------
# Models


def model_f1(
    r,
    g,
    b,
    theta_1,
    theta_2,
    center_x,
    center_y,
    radius,
    offset,
    image_size,
):
    image = np.ones((3, image_size, image_size))
    theta_1, theta_2 = np.deg2rad(theta_1), np.deg2rad(theta_2)
    # Color
    image = np.zeros((3, image_size, image_size))
    malus_factor = np.cos(theta_1 - theta_2) ** 2
    image[0] = r * malus_factor
    image[1] = g * malus_factor
    image[2] = b * malus_factor
    # Apply hexagon mask
    mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
    image *= mask
    return image.transpose((1, 2, 0)).astype(int)


def model_f2(
    r,
    g,
    b,
    theta_1,
    theta_2,
    S,
    w_r,
    w_g,
    w_b,
    exposure,
    center_x,
    center_y,
    radius,
    offset,
    image_size,
):
    # Transform parameters
    theta_1, theta_2 = np.deg2rad(theta_1), np.deg2rad(theta_2)
    r /= 255
    g /= 255
    b /= 255
    # Color
    malus_factor = np.cos(theta_1 - theta_2) ** 2
    W = np.diag([w_r, w_g, w_b])
    color = exposure * W @ S @ np.array([[r, g, b]]).T * malus_factor
    color = np.minimum(1, color)
    # Produce the image
    image = np.ones((3, image_size, image_size))
    image[0] *= color[0]
    image[1] *= color[1]
    image[2] *= color[2]
    mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
    image *= mask
    return image.transpose((1, 2, 0))


def model_f3(
    r,
    g,
    b,
    theta_1,
    theta_2,
    S,
    w_r,
    w_g,
    w_b,
    exposure,
    Tp,
    Tc,
    center_x,
    center_y,
    radius,
    offset,
    image_size,
):
    # Transform parameters
    theta_1, theta_2 = np.deg2rad(theta_1), np.deg2rad(theta_2)
    r /= 255
    g /= 255
    b /= 255
    # Color
    malus_factor = (Tp - Tc) * np.cos(theta_1 - theta_2) ** 2 + Tc
    W = np.diag([w_r, w_g, w_b])
    color = exposure * W @ S @ np.array([[r, g, b]]).T * malus_factor
    color = np.minimum(1, color)
    # Produce the image
    image = np.ones((3, image_size, image_size))
    image[0] *= color[0]
    image[1] *= color[1]
    image[2] *= color[2]
    mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
    image *= mask
    return image.transpose((1, 2, 0))



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


"""This module contains helper functions, e.g., to download and
extract .zip files and check their MD5 checksums.

"""

import numpy as np
import zipfile
import requests
from pathlib import Path
import hashlib
from tqdm import tqdm

# --------------------------------------------------------------------
# Functions to download, extract and verify datasets


def download_and_extract(
    url,
    root,
    checksum=None,
):
    """
    Parameters
    ----------
    url : string
        The download URL.
    root : string
        The path to the directory on the local computer, where the file will be downloaded and extrated.
    checksum : string or NoneType, default=None
        The expected MD5 checksum of the downloaded file, which will
        be checked against its actual checksum. If `None`, the
        checksum is not checked.

    Returns
    -------
    None

    """
    local_zipfile = "causal_chamber_" + hashlib.md5(url.encode()).hexdigest() + ".zip"
    zip_path = Path(root, local_zipfile)
    # Download
    _download(url, zip_path)
    # Verify
    if checksum is not None:
        print("  Verifying checksum...", end="")
        md5 = _compute_md5(zip_path, checksum)
        if checksum != md5:
            raise Exception(
                f'Checksum does not match!\n  expected: "{checksum}"\n  computed: "{md5}"'
            )
        else:
            print(" done.")
    # Extract
    print(f'  Extracting zip-file contents to "{root}"...', end="")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(root)
    print(" done.")


def _download(url, output_path):
    """Function to actually download the file from the given URL into the given output_path."""
    print(f'Downloading dataset from "{url}" into "{output_path}"\n')
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(output_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print("ERROR, something went wrong")


def _unzip(path, output_dir):
    """
    Unzips the file at the given path into the given output_dir.
    """
    print(f'  Extracting "{path}" to "{output_dir}"')
    with zipfile.ZipFile(path, "r") as zip_ref:
        zip_ref.extractall(output_dir)


def _compute_md5(path, checksum):
    """Verify the MD5 checksum of a file at the given path."""
    hasher = hashlib.md5()
    with open(path, "rb") as f:
        data = f.read()
        hasher.update(data)
    return hasher.hexdigest()


# --------------------------------------------------------------------
# Functions used by experiment protocol generators


def set_and_measure(
    value_ranges,
    n_interventions,
    n_msr,
    repeat=False,
    wait_before_set=0,
    wait_before_msr=0,
    random_state=42,
):
    """Generates a sequence of SET and MSR instructions with optional WAIT
    instructions between them, where the set values, measurement sizes and
    wait times are randomly selected.

    Parameters
    ----------
    value_ranges : dict
        A dictionary where each key is an intervention target to set,
        and each value is a set of possible values it can take.

    n_interventions : int
        The number of intervention cycles
        to perform. Each intervention cycle consists of setting
        values (SET), waiting (WAIT), and measuring (MSR).

    n_msr : int or tuple
        The number of measurements to take in each measurement
        phase. If a tuple, it defines a range of possible numbers to
        randomly sample from.

    repeat : bool, optional
        If True, allows the repetition of the same value for
        subsequent interventions for the same target. If False, each
        intervention for a target must have a different value from the
        last. Default is False.

    wait_before_set : int or tuple, optional
        The waiting time (in milliseconds) before setting values in
        each intervention. If a tuple, it defines a range of possible
        wait times to randomly sample from. Default is 0, meaning no
        WAIT instruction is produced.

    wait_before_msr : int or tuple, optional
        The waiting time (in milliseconds) before measuring after each
        intervention. If a tuple, it defines a range of possible wait
        times to randomly sample from. Default is 0, meaning no WAIT
        instruction is produced.

    random_state : int, optional
        A seed value for the random number generator to ensure
        reproducibility. Default is 42.

    Returns
    -------
    list of str
        A list of strings, each representing an instruction.

    Notes
    -----

    - The function internally uses a random number generator for
      selecting wait times, values to set, and the number of
      measurements.
    - The 'SET' instructions are generated based on the `value_ranges`
      and the `repeat` parameter.
    - The function ensures that each target is set exactly once in
      each intervention cycle.

    """
    # Function implementation
    last_value = dict((target, set()) for target, _ in value_ranges.items())
    instructions = []
    rng = np.random.default_rng(random_state)

    def sample(values):
        if isinstance(values, tuple):
            return rng.integers(values[0], values[1], endpoint=True)
        else:
            return values

    for i in range(n_interventions):
        # WAIT instruction, before measuring
        wait = sample(wait_before_set)
        if wait > 0:
            instructions.append(f"WAIT,{wait}")
        # SET instruction
        for target, values in value_ranges.items():
            possible_values = (
                list(values) if repeat else list(values - last_value[target])
            )
            value = rng.choice(possible_values)
            instructions.append(f"SET,{target},{value}")
            last_value[target] = {value}
        # WAIT instruction, before measuring
        wait = sample(wait_before_msr)
        if wait > 0:
            instructions.append(f"WAIT,{wait}")
        # MSR instruction
        n_measurements = sample(n_msr)
        instructions.append(f"MSR,{n_measurements},0")
    return instructions



__all__ = ["datasets", "utils", "ground_truth", "simulators"]

from . import datasets
from . import utils
from . import ground_truth
from . import simulators

citation = """
﻿@article{gamella2025chamber,
  author={Gamella, Juan L. and Peters, Jonas and B{\"u}hlmann, Peter},
  title={Causal chambers as a real-world physical testbed for {AI} methodology},
  journal={Nature Machine Intelligence},
  doi={10.1038/s42256-024-00964-x},
  year={2025},
}
"""

print(
    f"If you use our datasets or simulators for your work please consider citing:\n{citation}"
)



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import causalchamber.datasets.utils as utils
from pathlib import Path
import pandas as pd
import yaml
from PIL import Image
import numpy as np
import os
import requests

"""
This module defines the functions to download datasets and load them as pandas dataframes or numpy arrays.
"""

# --------------------------------------------------------------------
# Download the list of available datasets when this module is imported
DIRECTORY_URL = (
    "https://causalchamber.s3.eu-central-1.amazonaws.com/downloadables/directory.yaml"
)

print(f"\nFetching list of available datasets from {DIRECTORY_URL} ...", end="")

r = requests.get(DIRECTORY_URL)
if r.status_code != 200:
    raise ConnectionError(f'Unexpected HTTP status code "{r.status_code}: {r}".')
directory = yaml.load(r.content, Loader=yaml.Loader)
print(" done.")


def list_available():
    available_datasets = directory["datasets"].keys()
    print(f"Available datasets (last changes on {directory['last_updated']}):\n")
    for d in available_datasets:
        print(f"  {d}")
    print(
        "\nVisit https://causalchamber.org for a detailed description of each dataset."
    )
    return list(available_datasets)


# --------------------------------------------------------------------
# Base Dataset and Experiment classes


class Dataset:
    def __init__(self, name, root, download=True):
        available_datasets = directory["datasets"].keys()
        if name not in available_datasets:
            string = ""
            for d in available_datasets:
                string += '  "' + d + '"\n'
            raise ValueError(
                f'Dataset "{name}" is not available. Available datasets:\n{string}'
            )
        # Store attributes
        self.name = name
        self.root = root
        self.image = directory["datasets"][name]["image"]
        self.chamber = directory["datasets"][name]["chamber"]
        # Check if dataset has already been downloaded to root
        dataset_dir = Path(self.root, self.name)
        if os.path.isdir(dataset_dir):
            print(f'Dataset {self.name} found in "{dataset_dir}".')
        else:
            if download:
                # Download, verify and extract
                self.url = directory["datasets"][name]["url"]
                self.checksum = directory["datasets"][name]["md5"]
                utils.download_and_extract(self.url, self.root, self.checksum)
            else:
                raise FileNotFoundError(
                    f'Could not find dataset directory "{dataset_dir}". Set download=True or choose another root directory (root).'
                )
        # Load available experiments
        #   If not an image dataset, experiments are just .csv files in the dataset directory
        #   If an image dataset, each experiment is a folder
        #   containing the .csv file with measurements and a subfolder
        #   with the images

        if self.image:
            experiment_names = [p.name for p in Path(self.root).glob(f"{self.name}/*")]
            csv_paths = [
                Path(self.root, self.name, e, f"{e}.csv") for e in experiment_names
            ]
            experiments = [ImageExperiment(self.name, path) for path in csv_paths]
        else:
            csv_paths = [p for p in Path(self.root).glob(f"{self.name}/*.csv")]
            experiments = [Experiment(self.name, path) for path in csv_paths]
        # Store experiment dictionary
        assert len(experiments) > 0
        self.__experiments = dict((e.name, e) for e in experiments)

    def available_experiments(self):
        """Return the list of available experiments in this dataset."""
        return list(self.__experiments.keys())

    def get_experiment(self, name):
        """Return a particular experiment given its name (see Experiment or ImageExperiment classes)."""
        return self.__experiments[name]


class Experiment:
    def __init__(self, dataset_name, csv_path):
        self.dataset = dataset_name
        self.csv_path = csv_path
        self.name = csv_path.stem
        self.columns = pd.read_csv(self.csv_path, nrows=0).columns.tolist()

    def as_pandas_dataframe(self):
        """Returns a pandas dataframe with the experiment data (excl. images)"""
        return pd.read_csv(self.csv_path)

    def as_image_array(self):
        raise NotImplementedError("This is not an image dataset!")


class ImageExperiment(Experiment):
    def __init__(self, dataset_name, csv_path):
        super().__init__(dataset_name, csv_path)
        self.image_folders = {}
        for path in [p for p in Path(csv_path.parents[0]).glob("images_*")]:
            size = path.stem.split("_")[1]
            self.image_folders[size] = path

    def available_sizes(self):
        """Returns the available image sizes. Size is given in pixels, where `full` means `2000x2000 px'."""
        return list(self.image_folders.keys())

    def as_image_array(self, size):
        """Returns a numpy array with all the images along the first dimension (axis-0)"""
        if size not in self.image_folders.keys():
            raise ValueError(
                f" Size {size} not available; available image sizes: {list(self.image_folders.keys())}."
            )
        image_filenames = pd.read_csv(self.csv_path).image_file
        image_folder = self.image_folders[size]
        image_paths = [Path(image_folder, f) for f in image_filenames]
        return np.array([np.array(Image.open(f)) for f in image_paths])



#!/usr/bin/env python
# coding: utf-8

# # **Tutorial**: Using the simulator `lt.Deterministic`
# 
# [<<< Back to the simulator index](https://github.com/juangamella/causal-chamber-package/tree/main/causalchamber/simulators)
# 
# Other links:
# - [Derivation of the simulators](TODO) -> TODO
# - [Simulators source code](https://github.com/juangamella/causal-chamber-package/blob/main/causalchamber/simulators/lt/sensors/main.py)
# 
# If you use the simulator in your work, please consider citing:
# 
# ```
# TODO
# ```

# In[1]:


import sys
sys.path = ['../../..'] + sys.path

# %load_ext autoreload
# %autoreload 2


# ## Dependencies

# Besides the `causalchamber` package, the notebook also uses `matplotlib`, `numpy` and `pandas` (specified in `tutorials/requirements.txt`).

# In[2]:


import causalchamber
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from matplotlib import gridspec


# ## Initializing the simulator
# 
# First, import and initialize the simulator with its parameters.

# In[3]:


import causalchamber.simulators.lt as lt


# In[4]:


# Parameters fit to data on a separate calibration dataset
parameters = {
    "S": np.array(
        [[4.34165202, 2.8594438, 1.91565691], [0.89410569, 1.58679242, 2.68477163]]
    ),
    "d1": np.array(0.1),
    "d2": np.array(0.17),
    "d3": np.array(0.25),
    "Ts": np.array([0.42685688, 0.41231782, 0.46652526]),
    "Tp": np.array([0.2909342, 0.29624314, 0.30526271]),
    "Tc": np.array([0.00431032, 0.01447611, 0.02041075]),
    "Q": np.array([0.12834946, 0.12870338, 0.13078523]),
    "C0": np.array(58.13307318503893),
    "A": np.array(1.42261211),
    "a1": np.array(506.8049747171294),
    "a2": np.array(510.98343062774353),
}

# Initialize simulator
simulator = lt.Deterministic(**parameters) # store pre-trained model in the current directory


# ## Running the simulator

# To simulate, call the function `.simulate_from_inputs` with a dataframe containing the simulator inputs as columns. There can be additional columns, which the simulator will ignore.
# 
# The names of the necessary columns can be obtained as:

# In[5]:


simulator.inputs_names


# ### Example 1: Random inputs
# 
# Let's simulate the sensor outputs for random values of the light-source color (`red, green, blue`) while keeping the other inputs (i.e., the sensor parameters `diode_*, t_*, v_*`) constant.

# In[6]:


rng = np.random.default_rng(7); N = 100

inputs = pd.DataFrame(
    {
        "red": rng.uniform(0, 255, size=N),
        "green": rng.uniform(0, 255, size=N),
        "blue": rng.uniform(0, 255, size=N),
        "pol_1": rng.uniform(-180, 180, size=N),
        "pol_2": rng.uniform(-180, 180, size=N),
        "diode_ir_1": 2,
        "diode_ir_2": 2,
        "diode_ir_3": 2,
        "diode_vis_1": 1,
        "diode_vis_2": 1,
        "diode_vis_3": 1,
        "t_ir_1": 3,
        "t_ir_2": 3,
        "t_ir_3": 3,
        "t_vis_1": 3,
        "t_vis_2": 3,
        "t_vis_3": 3,
        "v_c": 5,
        "v_angle_1": 5,
        "v_angle_2": 5,
    }
)
outputs = simulator.simulate_from_inputs(inputs)


# You can transform the outputs into a dataframe like this:

# In[7]:


outputs = pd.DataFrame(dict(list(zip(simulator.outputs_names, outputs))))
outputs.iloc[1:5]


# Plot the results, e.g., response of the light sensors as a function of the inputs `red, green, blue, pol_1`and `pol_2`.

# In[8]:


# Color represents the light-source color
color = inputs[['red', 'green', 'blue']].values / 255
# Size represents the scaling effect of the polarizers (Malus' law)
size = 5 + 30 * np.cos(np.deg2rad(inputs.pol_1 - inputs.pol_2))**2
# x-axis is the average of the light-source setting
x = inputs[['red', 'green', 'blue']].values.mean(axis=1)

plt.figure(figsize=(12,4))
plt.subplot(131)
plt.scatter(x, outputs.ir_1, c=color, s=size)
plt.title("ir_1")
plt.xlabel("(red + green + blue) / 3")
plt.subplot(132)
plt.scatter(x, outputs.ir_2, c=color, s=size)
plt.title("ir_2")
plt.xlabel("(red + green + blue) / 3")
plt.subplot(133)
plt.scatter(x, outputs.ir_3, c=color, s=size)
plt.title("ir_3")
plt.xlabel("(red + green + blue) / 3")


# In the plot, the color of each point represents the light-source color (red, green, blue) and the size the scaling effect of the polarizers (Malus' law).
# 
# Only the third sensor, which is placed behind both polarizers, is affected by their relative positions.

# ### Example 2: Compare to real measurements

# We download a dataset of real measurements using the functionality provided in the `causalchamber` package. The downloaded dataset is stored in the path given by `root=` below.
# 
# For our experiment we will use the `lt_crl_benchmark_v1` dataset, which is described [here](TODO).

# In[9]:


# Download the dataset
dataset = causalchamber.datasets.Dataset('lt_crl_benchmark_v1', root='./', download=True)

# Load the data (inputs, measurements and images) from one of the experiments
experiment = dataset.get_experiment(name='uniform')
observations = experiment.as_pandas_dataframe()


# We can directly use the `observations` dataframe as input to the simulator, since it contains the necessary columns. The additional columns are ignored by the simulator.

# In[10]:


inputs = observations
outputs = simulator.simulate_from_inputs(inputs)

# transform into a dataframe
outputs = pd.DataFrame(dict(list(zip(simulator.outputs_names, outputs))))


# **Plot the results**: real vs. simulated measurements for each sensor
# 
# To show the latex symbol for each sensor, we can use the function `causalchamber.ground_truth.latex_name`.

# In[11]:


from causalchamber.ground_truth import latex_name


# In[12]:


plt.figure(figsize=(6,6), dpi=150)
gs = gridspec.GridSpec(3, 3, wspace=0.1, hspace=0.1)

# real data
real_df = observations

# synthetic data
synth_df = outputs

# Compare observations    
for i,var in enumerate(simulator.outputs_names):
    plt.subplot(gs[i])
    ax = plt.gca()
    truth = real_df[var]
    
    # Again, color for each plotted point is given by the corresponding light-source color
    c = real_df[['red', 'green', 'blue']].values / 256
    
    # Normalize real and synthetic data for easier plotting
    mmin = truth.min()
    mmax = truth.max()
    truth = (truth - mmin) / (mmax - mmin)
    pred = (synth_df[var] - mmin) / (mmax - mmin)
    
    # Plot
    ax.scatter(truth, pred, marker='.', s=30, c=c, edgecolor='#2a2a2a', linewidth=0.2)        

    # Add sensor name
    ax.text(0.1, 0.9, latex_name(var, enclose=True) + " - " + str(var), ha="left", va="top", rotation=0, size=9, transform=ax.transAxes)    

    # Set axes limits
    lim = [-0.05,1.05]
    ax.plot(lim, lim, color='black', zorder=-10, linewidth=1, linestyle="-")#(0, (2, 1)))
    plt.ylim(lim)
    plt.xlim(lim)

    # Add ticks and axis labels
    plt.xticks([0,0.5,1], [])
    plt.yticks([0,0.5,1], [])
    if i >= 6:
        plt.xlabel("Real measurement", size=8)
    if i in [0,3,6]:
        plt.ylabel("Synth. measurement", size=8)


# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# # **Tutorial**: Using the simulators `wt.ModelA1`, `wt.ModelA2` and `wt.ModelB1`
# 
# [<<< Back to the simulator index](https://github.com/juangamella/causal-chamber-package/tree/main/causalchamber/simulators)
# 
# Other links:
# - [Derivation of the simulators](https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332)
# - [Simulators source code](https://github.com/juangamella/causal-chamber-package/blob/main/causalchamber/simulators/wt/main.py)
# 
# If you use the simulators in your work, please consider citing:
# 
# ```
# ﻿@article{gamella2025chamber,
#   author={Gamella, Juan L. and Peters, Jonas and B{\"u}hlmann, Peter},
#   title={Causal chambers as a real-world physical testbed for {AI} methodology},
#   journal={Nature Machine Intelligence},
#   doi={10.1038/s42256-024-00964-x},
#   year={2025},
# }
# ```

# In[1]:


import sys
sys.path = ['../../..'] + sys.path

# %load_ext autoreload
# %autoreload 2


# ## Dependencies

# Besides the `causalchamber` package, the notebook also uses `matplotlib`, `numpy` and `pandas` (specified in `tutorials/requirements.txt`).

# In[2]:


import causalchamber
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from matplotlib import gridspec


# ## Initializing the simulators
# 
# We take the simulator parameters from calibration experiments and the technical specifications of the fans. See [this notebook](https://github.com/juangamella/causal-chamber-paper/blob/main/case_studies/mechanistic_models.ipynb) for the complete process.

# In[3]:


# For models A1 and B1
C_MIN = 0.166
C_MAX = 0.27
L_MIN = 0.1
OMEGA_MAX = 3000

# For model A2
I = 3.481e-05
T = 0.05
K = 5.268701549401565e-08 # This is called "C" in the notebook linked above

#   Torque function (also a parameter for model A2)
def tau(load, C_min=C_MIN, C_max=C_MAX, L_min=L_MIN, T=T):
    load = np.atleast_1d(load)
    torques = T * (C_min + np.maximum(L_min, load) ** 3 * (C_max - C_min) - C_min)
    torques[load == 0] = 0
    return torques if len(load) > 1 else torques[0]


# In[4]:


import causalchamber.simulators.wt as wt

sim_a1 = wt.ModelA1(L_min = L_MIN,
                    omega_max = OMEGA_MAX)

sim_a2 = wt.ModelA2(I = I,
                    tau = tau,
                    K = K,
                    omega_0 = L_MIN * OMEGA_MAX) # We assume the initial velocity is the idle speed of the fans

sim_b1 = wt.ModelB1(C_min=C_MIN,
                    C_max=C_MAX,
                    L_min=L_MIN)


# ## Running the simulators

# To simulate, call the function `.simulate_from_inputs` with a dataframe containing the simulator inputs as columns. There can be additional columns, which the simulator will ignore.
# 
# The names of the necessary columns can be obtained by calling:

# In[5]:


sim_a1.inputs_names


# In[6]:


sim_a2.inputs_names


# In[7]:


sim_b1.inputs_names


# ### Example 1: Random inputs
# 
# #### Model A1 and B1
# 
# We can use models A1 and B1 to simulate the steady-state fan speed and current (respectively). As inputs, let us take random fan loads from the valid range [0,1].

# In[8]:


rng = np.random.default_rng(7); N = 100
inputs = pd.DataFrame({'load': rng.uniform(0,1,size=N)})

fan_speed = sim_a1.simulate_from_inputs(inputs)
fan_current = sim_b1.simulate_from_inputs(inputs)


# In[9]:


plt.figure(figsize=(8,4))
plt.subplot(1,2,1)
plt.scatter(inputs.load, fan_speed)
plt.title("Fan speed (rpm)")
plt.xlabel("Fan load")

plt.subplot(1,2,2)
plt.scatter(inputs.load, fan_current)
plt.title("Fan current (Amps)")
plt.xlabel("Fan load")


# #### Model A2
# 
# The simulator models the dynamics of the fan speed, given a time series of the load and the timestamps at which measurements happen.
# 
# To keep it simple, let us simulate what happens when we turn the load on (`load=1`) and off (`load=0`) in a short impulse, and take measurements every 0.1 seconds.

# In[10]:


inputs = pd.DataFrame({'load': np.array([0] * 30 + [1] * 20 + [0] * 50), 'timestamp': np.linspace(0,10,100)})
fan_speed = sim_a2.simulate_from_inputs(inputs)


# In[11]:


plt.figure()
ax1 = plt.gca()

c1,c2 = 'tab:blue', 'tab:purple'

ax1.set_xlabel('time (s)')
ax1.set_ylabel('fan\nload', color=c1, rotation=0, ha="right")
ax1.plot(inputs.timestamp, inputs.load, color=c1, linestyle='--')
ax1.tick_params(axis='y', labelcolor=c1)

ax2 = ax1.twinx()
ax2.set_ylabel('fan\nspeed\n(rpm)', color=c2, rotation=0, ha="left")
ax2.plot(inputs.timestamp, fan_speed, color=c2)
ax2.tick_params(axis='y', labelcolor=c2)


# ### Example 2: Comparing to real measurements

# We download a dataset of real measurements using the functionality provided in the `causalchamber` package. The downloaded dataset is stored in the path given by `root=` below.
# 
# For our experiment we will use the `wt_test_v1` dataset, which is described [here](https://github.com/juangamella/causal-chamber/tree/main/datasets/wt_test_v1).

# In[12]:


# Download the dataset
dataset = causalchamber.datasets.Dataset('wt_test_v1', root='./', download=True)

# Load the "steps" experiment
experiment = dataset.get_experiment(name='steps')
observations = experiment.as_pandas_dataframe().iloc[0:100]


# We will compare our simulator with the behaviour of the intake fan in the above experiment.
# 
# First, let us create the inputs for all three simulator (we need `load` and `timestamp`). The simulators A1 and B1 will ignore the `timestamp` column.

# In[13]:


inputs = pd.DataFrame({'load': observations.load_in,
                       'timestamp': observations.timestamp})

output_a1 = sim_a1.simulate_from_inputs(inputs)
output_b1 = sim_b1.simulate_from_inputs(inputs)
output_a2 = sim_a2.simulate_from_inputs(inputs)


# **Plot the results**: real vs. simulated measurements of the fan speed and current

# In[14]:


plt.figure(figsize=(16,4))

# x-axis is elapsed time since beginning of experiment
t = inputs.timestamp - inputs.timestamp.iloc[0]

# Plot 1: load
plt.subplot(141)
plt.title("Inputs")
plt.plot(t, inputs.load, label="Load")
plt.legend()
plt.xlabel("time (s)")

# Plot 2: Model A1 vs. fan speed
plt.subplot(142)
plt.title("Model A1")
plt.plot(t, output_a1, label="Model A1")
plt.plot(t, observations.rpm_in, label="Real fan speed")
plt.legend()
plt.xlabel("time (s)")

# Plot 3: Model A3 vs. fan speed
plt.subplot(143)
plt.title("Model A2")
plt.plot(t, output_a2, label="Model A2")
plt.plot(t, observations.rpm_in, label="Real fan speed")
plt.legend()
plt.xlabel("time (s)")

# Plot 4: Model B1 vs. fan current
plt.subplot(144)
# See Table I in Appendix II of [1] to see how the current in Amps is calculated
current_amps = observations.current_in / 1023 * 2.5 * 1.16 / 5 
plt.title("Model B1")
plt.plot(t, output_b1, label="Model B1")
plt.plot(t, current_amps, label="Real fan current")
plt.legend()
plt.xlabel("time (s)")


# In[ ]:







# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest

import pandas as pd

col_numbers = {
    "lt": {"standard": 46, "camera": 52},
    "wt": {
        "standard": 37,
        "pressure-control": 45,
    },
}


class DownloadTests(unittest.TestCase):
    def test_workflow(self):
        """Test that all datasets listed in the directory are downloadable and
        their experiments accessible. Check that the number of columns is
        consistent across experiments and with the corresponding chamber
        configuration."""
        import causalchamber.datasets

        available = causalchamber.datasets.list_available()
        for name in available:
            print(f"dataset: {name}")
            dataset = causalchamber.datasets.Dataset(
                name=name, root="/tmp", download=True
            )
            experiments = dataset.available_experiments()
            print(experiments)
            prev_columns = None
            prev_config = None
            for e in experiments:
                experiment = dataset.get_experiment(e)
                df = experiment.as_pandas_dataframe()
                print(f"  experiment: {e}")
                print(f"    {len(df)} observattions")
                print(f"    columns {len(df.columns)}")
                # Check column names and chamber config are consistent across the experiments of this dataset.
                if prev_columns is None:
                    for c in df.columns:
                        print("     ", c)
                    prev_columns = df.columns
                    config = pd.unique(df.config)
                    self.assertEqual(len(config), 1)
                    prev_config = config[0]
                    self.assertEqual(
                        len(df.columns), col_numbers[dataset.chamber][config[0]]
                    )
                else:
                    self.assertTrue((df.columns == prev_columns).all())
                    config = pd.unique(df.config)
                    self.assertEqual(len(config), 1)
                    self.assertEqual(config[0], prev_config)
                    self.assertEqual(
                        len(df.columns), col_numbers[dataset.chamber][config[0]]
                    )
                # If an image dataset, check that all available sizes can be loaded
                if dataset.image:
                    sizes = experiment.available_sizes()
                    print(f"    available image sizes: {sorted(sizes)}")
                    for size in sorted(sizes):
                        print(f"      loading size {size}")
                        images = experiment.as_image_array(sorted(sizes)[0])
                        self.assertEqual(len(df), len(images))
                        print("      ok.")



__all__ = []
from .sensors.main import Deterministic
from .image.models_f import ModelF1, ModelF2, ModelF3

# Optional import for DecoderSimple (needs optional dependencies listed under [torch]
try:
    from .image.decoder import DecoderSimple
except ImportError as e:
    print(
        "WARNING: PyTorch (torch + torchvision) is required to run the simulator lt.DecoderSimple. Simulator was not imported."
    )



# MIT License

# Copyright (c) 2025 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from causalchamber.simulators import Simulator
import requests
from pathlib import Path
import hashlib
import os
import numpy as np
import datetime
from tqdm import tqdm
import torch
import torch.nn as nn
import torchvision.transforms as transforms

# import wandb

# --------------------------------------------------------------------
# Public API

# Download URL for the pre-trained model
TORCH_MODEL_URL = "https://causalchamber.s3.eu-central-1.amazonaws.com/downloadables/sparkling-elevator-10.pkl"


class DecoderSimple(Simulator):
    inputs_names = [
        "red",
        "green",
        "blue",
        "pol_1",
        "pol_2",
    ]
    outputs_names = ["image"]

    def __init__(
        self,
        torch_model_url=TORCH_MODEL_URL,
        root="./",
        device="cpu",
        download=True,
    ):
        """
        Initializes the simulator by downloading and loading a pre-trained Torch model.

        The function checks whether the model has already been downloaded to the `root` directory.
        If the file is not found and `download` is set to True, it downloads the model from
        the given `torch_model_url`. Otherwise, it raises a `FileNotFoundError`. The model
        is then loaded onto the specified `device`, e.g., `cpu` or `cuda`.

        Parameters
        ----------
        torch_model_url : str, optional
            URL of the pre-trained Torch model (default: `TORCH_MODEL_URL`).
        root : str, optional
            Path to the directory where the model should be stored (default: "./").
        device : str, optional
            Device to load the model on ("cpu" or "cuda") (default: "cpu").
        download : bool, optional
            Whether to download the model if not found locally (default: True).

        Raises
        ------
        FileNotFoundError
            If the model is not found in the specified `root` directory and `download` is set to False.

        Notes
        -----
        - The model is stored with a unique filename derived from the hash of `torch_model_url`.
        - The `DecoderNetwork` class is used to load the Torch model (see below).
        - If the model is already available, it skips downloading and directly loads the model.
        """
        super(DecoderSimple, self).__init__()
        # Check if model has been already downloaded
        local_file = (
            "torch_model_" + hashlib.md5(torch_model_url.encode()).hexdigest() + ".pkl"
        )
        download_path = Path(root, local_file)
        if os.path.exists(download_path):
            print(f'Pre-trained model "{local_file}" found in "{root}".')
        else:
            if download:
                _download(torch_model_url, download_path)
            else:
                raise FileNotFoundError(
                    f'Could not find model in directory "{root}". Set download=True or choose another root directory (root).'
                )

        # Load the torch model
        # device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        self.torch_model = DecoderNetwork(device=torch.device(device))
        self.torch_model.load(download_path)
        print("Model loaded.")

    def _simulate(
        self,
        red,
        green,
        blue,
        pol_1,
        pol_2,
    ):
        """Produces synthetic images given values for the light-source color
        (`red,green,blue`) and polarizer positions (`pol_1, pol_2`).

        Parameters
        ----------
        red, green, blue : array_like
            1D arrays representing the brightness settings of the red, green, and blue LEDs
            respectively, each in the range [0, 255].
        pol_1, pol_2 : array_like
            1D arrays representing the positions of the tunnel's polarizers in degrees, in
            the range [-180, 180].

        Returns
        -------
        outputs : numpy.ndarray
            The array containing the images with shape (N,64,64,3)
            where N is the number of produced images (equal to the
            length of the input arrays red, green, ...). The values of
            the array are between 0 and 1.

        """
        # Normalize the inputs (so they're between -1 and 1)
        inputs = np.array([red, green, blue, pol_1, pol_2]).T
        # print(inputs.shape)
        inputs = (inputs - np.array([128, 128, 128, 0, 0])) / np.array(
            [128, 128, 128, 180, 180]
        )
        try:
            inputs = torch.tensor(inputs, dtype=torch.float32)
        except TypeError:
            inputs = torch.tensor(
                np.asarray(inputs, dtype=np.float32), dtype=torch.float32
            )
        outputs = self.torch_model(inputs).detach().cpu().numpy()
        # Unnormalize images
        outputs = (outputs + 1) / 2
        # Clip to valid image range [0,1]
        outputs = np.maximum(outputs, 0)
        outputs = np.minimum(outputs, 1)
        # Set color axis last (accounting for possible extra batch axis)
        axes = (0, 2, 3, 1) if outputs.ndim == 4 else (1, 2, 0)
        return np.transpose(outputs, axes=axes)


def _download(url, output_path):
    """Function to download the file from the given URL into the given output_path."""
    print(f'Downloading model from "{url}" into "{output_path}"\n')
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(output_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print("ERROR, something went wrong")


# --------------------------------------------------------------------
# Underlying PyTorch Model
#   for more details: see Appendix C of "Sanity Checking Causal
#   Representation Learning on a Simple Real-World System (2025) by
#   Juan L. Gamella*, Simom Bing* and Jakob Runge)


class TorchImageDataset(torch.utils.data.Dataset):
    """
    To load the image datasets used in training the DecoderNetwork below.
    """

    def __init__(self, images, labels, mean_labels=None, std_labels=None):
        # Normalize labels
        self.mean_labels = (
            labels.mean(axis=0, keepdims=True) if mean_labels is None else mean_labels
        )
        self.std_labels = (
            labels.std(axis=0, keepdims=True) if std_labels is None else std_labels
        )
        self.n = len(labels)
        self.labels = torch.tensor(
            (labels - self.mean_labels) / self.std_labels, dtype=torch.float32
        )

        # Process images: tensors with pixel values between [-1,1]
        transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )
        self.images = [torch.unsqueeze(transform(im), 0) for im in images]
        # torch.unsqueeze(torch.flatten(transform(im)), 0) for im in images
        self.images = torch.cat(self.images)  # Concatenate images into a single tensor

    def __len__(self):
        return self.n

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        return self.labels[idx], self.images[idx]


class DecoderNetwork(nn.Module):
    """
    Torch model used to train/call the MLP used in the simulator.
    """

    def __init__(
        self,
        device="cuda",
    ):
        super().__init__()
        # Define layers
        self.image_size = 64
        hidden_dim = 4096
        self.net = nn.Sequential(
            # Input layer
            nn.Linear(5, hidden_dim),
            nn.ReLU(),
            # Hidden layer 1
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            # Hidden layer 2
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            # Output layer
            nn.Linear(hidden_dim, 3 * self.image_size * self.image_size),
        )
        # Set device
        self.device = device
        self.to(device)

    def forward(self, x):
        output = self.net(x.to(self.device))
        if x.dim() == 1:  # single tensor
            return output.view(3, self.image_size, self.image_size)
        else:  # account for batch size
            return output.view(output.shape[0], 3, self.image_size, self.image_size)

    def save(self, directory=None, path=None, tag=None):
        if directory is not None:
            timestamp = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
            path = (
                directory
                + f"model_{timestamp}"
                + ("" if tag is None else f"tag_{tag}")
                + ".pkl"
            )
        elif path is None:
            raise ValueError("Must provide either directory or path")
        torch.save(self.state_dict(), path)
        print(f'Saved model {self} to "{path}"')

    def load(self, path):
        # device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        self.load_state_dict(torch.load(path, map_location=torch.device("cpu")))

    def train(
        self,
        dataset_train,
        dataset_test,
        epochs=30,
        batch_size=2048,
        learning_rate=1e-3,
        weight_decay=1e-5,
        every=15,
    ):
        # Optional: log with weights & biases
        # wandb.init(
        #     # set the wandb project where this run will be logged
        #     project="chamber_models_decoder",
        #     # track hyperparameters and run metadata
        #     config={
        #         "tag": log_wandb,
        #         "learning_rate": learning_rate,
        #         "epochs": epochs,
        #         "batch_size": batch_size,
        #         "weight_decay": weight_decay,
        #         "device": self.device,
        #     },
        # )

        # Define loss and optimizer
        loss_function = torch.nn.MSELoss()
        optimizer = torch.optim.Adam
        opt_args = {"lr": learning_rate, "weight_decay": weight_decay}
        optimizer = optimizer(self.parameters(), **opt_args)

        # Dataloaders
        dataloader_args = {"shuffle": True, "num_workers": 0}
        train_dataloader = torch.utils.data.DataLoader(
            dataset_train, batch_size=batch_size, **dataloader_args
        )
        X_test, Y_test = next(
            iter(
                torch.utils.data.DataLoader(dataset_test, batch_size=len(dataset_test))
            )
        )

        # Training loop
        for epoch in tqdm(range(epochs)):
            # start = time.time()
            for i, (X, Y) in enumerate(train_dataloader):
                # Zero gradients
                optimizer.zero_grad()

                # Forward pass
                y_pred = self(X)
                loss = loss_function(y_pred, Y.to(self.device))

                # Zero gradients, backward pass, and update weights
                loss.backward()
                optimizer.step()

                # Compute test loss
                # test_loss = loss_function(self(X_test), Y_test.to(self.device)).item()
                # elapsed = time.time() - start
                # wandb.log({"train_loss": loss, "test_loss": test_loss})

        # Close wandb session
        # wandb.finish()



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module defines the variables and ground-truth edges/graphs
for each chamber configuration. The `latex_name` function allows you
to go from the column name of a variable to its latex expression used
in the original paper (TODO).

"""

import pandas as pd
import numpy as np

# --------------------------------------------------------------------
# Definition of variables and latex names

_variables_lt_standard = [
    "red",
    "green",
    "blue",
    "osr_c",
    "v_c",
    "current",
    "pol_1",
    "pol_2",
    "osr_angle_1",
    "osr_angle_2",
    "v_angle_1",
    "v_angle_2",
    "angle_1",
    "angle_2",
    "ir_1",
    "vis_1",
    "ir_2",
    "vis_2",
    "ir_3",
    "vis_3",
    "l_11",
    "l_12",
    "l_21",
    "l_22",
    "l_31",
    "l_32",
    "diode_ir_1",
    "diode_vis_1",
    "diode_ir_2",
    "diode_vis_2",
    "diode_ir_3",
    "diode_vis_3",
    "t_ir_1",
    "t_vis_1",
    "t_ir_2",
    "t_vis_2",
    "t_ir_3",
    "t_vis_3",
]

_variables_lt_camera = _variables_lt_standard + [
    "im",
    "aperture",
    "shutter_speed",
    "iso",
]

_variables_wt_standard = [
    "hatch",
    "pot_1",
    "pot_2",
    "osr_1",
    "osr_2",
    "osr_mic",
    "osr_in",
    "osr_out",
    "osr_upwind",
    "osr_downwind",
    "osr_ambient",
    "osr_intake",
    "v_1",
    "v_2",
    "v_mic",
    "v_in",
    "v_out",
    "load_in",
    "load_out",
    "current_in",
    "current_out",
    "res_in",
    "res_out",
    "rpm_in",
    "rpm_out",
    "pressure_upwind",
    "pressure_downwind",
    "pressure_ambient",
    "pressure_intake",
    "mic",
    "signal_1",
    "signal_2",
]

_variables_wt_pressure_control = _variables_wt_standard

_latex_names = {
    # Light tunnel variables
    "red": r"R",
    "green": r"G",
    "blue": r"B",
    "osr_c": r"O_C",
    "v_c": r"R_C",
    "current": r"\tilde{C}",
    "pol_1": r"\theta_1",
    "pol_2": r"\theta_2",
    "osr_angle_1": r"O_1",
    "osr_angle_2": r"O_2",
    "v_angle_1": r"R_1",
    "v_angle_2": r"R_2",
    "angle_1": r"\tilde{\theta}_1",
    "angle_2": r"\tilde{\theta}_2",
    "ir_1": r"\tilde{I}_1",
    "vis_1": r"\tilde{V}_1",
    "ir_2": r"\tilde{I}_2",
    "vis_2": r"\tilde{V}_2",
    "ir_3": r"\tilde{I}_3",
    "vis_3": r"\tilde{V}_3",
    "l_11": r"L_{11}",
    "l_12": r"L_{12}",
    "l_21": r"L_{21}",
    "l_22": r"L_{22}",
    "l_31": r"L_{31}",
    "l_32": r"L_{32}",
    "diode_ir_1": r"D^I_1",
    "diode_vis_1": r"D^V_1",
    "diode_ir_2": r"D^I_2",
    "diode_vis_2": r"D^V_2",
    "diode_ir_3": r"D^I_3",
    "diode_vis_3": r"D^V_3",
    "t_ir_1": r"T^I_1",
    "t_vis_1": r"T^V_1",
    "t_ir_2": r"T^I_2",
    "t_vis_2": r"T^V_2",
    "t_ir_3": r"T^I_3",
    "t_vis_3": r"T^V_3",
    "im": r"\tilde{\text{I}}\text{m}",
    "shutter_speed": r"T_\text{Im}",
    "aperture": r"\text{Ap}",
    "iso": r"\text{ISO}",
    # Wind tunnel variables
    "hatch": r"H",
    "pot_1": r"A_1",
    "pot_2": r"A_2",
    "osr_1": r"O_1",
    "osr_2": r"O_2",
    "osr_mic": r"O_M",
    "osr_in": r"O_\text{in}",
    "osr_out": r"O_\text{out}",
    "osr_upwind": r"O_\text{up}",
    "osr_downwind": r"O_\text{dw}",
    "osr_ambient": r"O_\text{amb}",
    "osr_intake": r"O_\text{int}",
    "v_1": r"R_1",
    "v_2": r"R_2",
    "v_mic": r"R_M",
    "v_in": r"R_\text{in}",
    "v_out": r"R_\text{out}",
    "load_in": r"L_\text{in}",
    "load_out": r"L_\text{out}",
    "current_in": r"\tilde{C}_\text{in}",
    "current_out": r"\tilde{C}_\text{out}",
    "res_in": r"T_\text{in}",
    "res_out": r"T_\text{out}",
    "rpm_in": r"\tilde{\omega}_\text{in}",
    "rpm_out": r"\tilde{\omega}_\text{out}",
    "pressure_upwind": r"\tilde{P}_\text{up}",
    "pressure_downwind": r"\tilde{P}_\text{dw}",
    "pressure_ambient": r"\tilde{P}_\text{amb}",
    "pressure_intake": r"\tilde{P}_\text{int}",
    "mic": r"\tilde{M}",
    "signal_1": r"\tilde{S}_1",
    "signal_2": r"\tilde{S}_2",
}


# --------------------------------------------------------------------
# Definition of ground truth edges


_edges_lt_standard = [
    ("red", "ir_1"),
    ("green", "ir_1"),
    ("blue", "ir_1"),
    ("red", "ir_2"),
    ("green", "ir_2"),
    ("blue", "ir_2"),
    ("red", "ir_3"),
    ("green", "ir_3"),
    ("blue", "ir_3"),
    ("red", "vis_1"),
    ("green", "vis_1"),
    ("blue", "vis_1"),
    ("red", "vis_2"),
    ("green", "vis_2"),
    ("blue", "vis_2"),
    ("red", "vis_3"),
    ("green", "vis_3"),
    ("blue", "vis_3"),
    ("red", "current"),
    ("green", "current"),
    ("blue", "current"),
    ("pol_1", "ir_3"),
    ("pol_2", "ir_3"),
    ("pol_1", "vis_3"),
    ("pol_2", "vis_3"),
    ("pol_1", "angle_1"),
    ("pol_2", "angle_2"),
    ("v_angle_1", "angle_1"),
    ("osr_angle_1", "angle_1"),
    ("v_angle_2", "angle_2"),
    ("osr_angle_2", "angle_2"),
    ("v_c", "current"),
    ("osr_c", "current"),
    ("l_11", "ir_1"),
    ("l_12", "ir_1"),
    ("l_11", "vis_1"),
    ("l_12", "vis_1"),
    ("t_ir_1", "ir_1"),
    ("diode_ir_1", "ir_1"),
    ("t_vis_1", "vis_1"),
    ("diode_vis_1", "vis_1"),
    ("l_21", "ir_2"),
    ("l_22", "ir_2"),
    ("l_21", "vis_2"),
    ("l_22", "vis_2"),
    ("t_ir_2", "ir_2"),
    ("diode_ir_2", "ir_2"),
    ("t_vis_2", "vis_2"),
    ("diode_vis_2", "vis_2"),
    ("l_31", "ir_3"),
    ("l_32", "ir_3"),
    ("l_31", "vis_3"),
    ("l_32", "vis_3"),
    ("t_ir_3", "ir_3"),
    ("diode_ir_3", "ir_3"),
    ("t_vis_3", "vis_3"),
    ("diode_vis_3", "vis_3"),
]

_edges_lt_camera = _edges_lt_standard + [
    ("pol_1", "im"),
    ("pol_2", "im"),
    ("red", "im"),
    ("green", "im"),
    ("blue", "im"),
    ("shutter_speed", "im"),
    ("aperture", "im"),
    ("iso", "im"),
]

_edges_wt_standard = [
    ("load_in", "rpm_in"),
    ("res_in", "rpm_in"),
    ("load_in", "rpm_out"),
    ("load_in", "current_in"),
    ("load_in", "current_out"),
    ("load_out", "rpm_in"),
    ("load_out", "rpm_out"),
    ("res_out", "rpm_out"),
    ("load_out", "current_out"),
    ("load_out", "current_in"),
    ("hatch", "rpm_in"),
    ("hatch", "rpm_out"),
    ("load_in", "pressure_intake"),
    ("hatch", "pressure_intake"),
    ("load_out", "pressure_intake"),
    ("osr_intake", "pressure_intake"),
    ("load_in", "pressure_upwind"),
    ("hatch", "pressure_upwind"),
    ("load_out", "pressure_upwind"),
    ("osr_upwind", "pressure_upwind"),
    ("load_in", "pressure_downwind"),
    ("hatch", "pressure_downwind"),
    ("load_out", "pressure_downwind"),
    ("osr_downwind", "pressure_downwind"),
    ("osr_ambient", "pressure_ambient"),
    ("osr_in", "current_in"),
    ("v_in", "current_in"),
    ("osr_out", "current_out"),
    ("v_out", "current_out"),
    ("pot_1", "signal_1"),
    ("osr_1", "signal_1"),
    ("v_1", "signal_1"),
    ("pot_1", "signal_2"),
    ("pot_2", "signal_2"),
    ("osr_2", "signal_2"),
    ("v_2", "signal_2"),
    ("pot_1", "mic"),
    ("load_in", "mic"),
    ("load_out", "mic"),
    ("hatch", "mic"),
    ("osr_mic", "mic"),
    ("v_mic", "mic"),
]

_edges_wt_pressure_control = _edges_wt_standard + [
    ("pressure_downwind", "load_in"),
    ("pressure_downwind", "load_out"),
]

# --------------------------------------------------------------------
# Public API of the module


def latex_name(var, enclose=True):
    """
    Obtain a variable's LaTeX name from its name as it appears in the columns of the .csv files in datasets.

    Parameters
    ----------
    var : str
        The variable name as it appears in dataset columns.
    enclose : bool, optional
        If True, encloses the LaTeX name in dollar signs. Default is True.

    Returns
    -------
    str
        The LaTeX-formatted name of the variable as given in the original paper.

    """
    if var not in _latex_names:
        return var
    else:
        name = _latex_names[var]
        return "$" + name + "$" if enclose else name


def variables(chamber, configuration):
    """
    Return the variables provided by the given chamber in the given configuration.

    Parameters
    ----------
    chamber : str
        The chamber e.g., 'lt' for the light tunnel, 'wt' for the wind tunnel.
    configuration : str
        The configuration of the chamber. e.g., 'standard', 'camera', 'pressure-control'.

    Returns
    -------
    list
        A list of variables relevant to the specified chamber and configuration.

    Raises
    ------
    ValueError
        If the chamber/configuration combination is unknown.

    >>> len(variables('lt', 'standard'))
    38
    >>> len(variables('lt', 'camera'))
    42
    >>> len(variables('wt', 'standard'))
    32
    >>> len(variables('wt', 'pressure-control'))
    32

    """
    if chamber == "lt" and configuration == "standard":
        return _variables_lt_standard
    if chamber == "lt" and configuration == "camera":
        return _variables_lt_camera
    elif chamber == "wt" and configuration == "standard":
        return _variables_wt_standard
    elif chamber == "wt" and configuration == "pressure-control":
        return _variables_wt_pressure_control
    else:
        raise ValueError(f"Unknown chamber/configuration: {chamber}/{configuration}")


def edges(chamber, configuration):
    """
    Return the edges in the ground truth graph of the given chamber in the given configuration.

    Parameters
    ----------
    chamber : str
        The chamber e.g., 'lt' for the light tunnel, 'wt' for the wind tunnel.
    configuration : str
        The configuration of the chamber. e.g., 'standard', 'camera', 'pressure-control'.

    Returns
    -------
    list of tuples
        A list of tuples representing the edges in the ground truth graph for the specified chamber and configuration. A tuple `(i,j)` denotes the edge `i -> j`.

    Raises
    ------
    ValueError
        If the chamber/configuration combination is unknown.

    >>> len(edges('lt', 'standard'))
    57
    >>> len(edges('lt', 'camera'))
    65
    >>> len(edges('wt', 'standard'))
    42
    >>> len(edges('wt', 'pressure-control'))
    44

    """
    if chamber == "lt" and configuration == "standard":
        return _edges_lt_standard
    if chamber == "lt" and configuration == "camera":
        return _edges_lt_camera
    elif chamber == "wt" and configuration == "standard":
        return _edges_wt_standard
    elif chamber == "wt" and configuration == "pressure-control":
        return _edges_wt_pressure_control
    else:
        raise ValueError(f"Unknown chamber/configuration: {chamber}/{configuration}")


def graph(chamber, configuration):
    """
    Return the adjacency matrix of the ground truth graph of the given chamber in the given configuration.

    Parameters
    ----------
    chamber : str
        The chamber e.g., 'lt' for the light tunnel, 'wt' for the wind tunnel.
    configuration : str
        The configuration of the chamber. g.g., 'standard', 'camera', 'pressure-control'.

    Returns
    -------
    DataFrame
        A pandas DataFrame representing the adjacency matrix of the graph, with variables as both index and columns. An entry `[from,to] != 1` denotes the edge `from -> to`.

    >>> int(graph('lt', 'standard').values.sum())
    57
    >>> int(graph('lt', 'camera').values.sum())
    65
    >>> int(graph('wt', 'standard').values.sum())
    42
    >>> int(graph('wt', 'pressure-control').values.sum())
    44

    """
    nodes = variables(chamber, configuration)
    p = len(nodes)
    idx = dict(zip(nodes, np.arange(p)))
    adjacency = np.zeros((p, p), dtype=int)
    for fro, to in edges(chamber, configuration):
        adjacency[idx[fro], idx[to]] = 1
    return pd.DataFrame(adjacency, index=nodes, columns=nodes)


# ----------------------------------------------------------------------
# Doctests
if __name__ == "__main__":
    import doctest

    # Check consistency between graphs and variables
    assert set(graph("lt", "standard").columns) == set(variables("lt", "standard"))
    assert set(graph("lt", "standard").index) == set(variables("lt", "standard"))

    assert set(graph("lt", "camera").columns) == set(variables("lt", "camera"))
    assert set(graph("lt", "camera").index) == set(variables("lt", "camera"))

    assert set(graph("wt", "standard").columns) == set(variables("wt", "standard"))
    assert set(graph("wt", "standard").index) == set(variables("wt", "standard"))

    assert set(graph("wt", "pressure-control").columns) == set(
        variables("wt", "pressure-control")
    )

    doctest.testmod(
        extraglobs={},
        verbose=True,
        optionflags=doctest.ELLIPSIS,
    )



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Simulators for the wind tunnel phenomena.
"""


import numpy as np
from causalchamber.simulators import Simulator

# --------------------------------------------------------------------
# Simulator classes


class ModelA1(Simulator):
    """Simulator of the steady-state fan speed given the load.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model A1 in Appendix IV.1 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    """

    # Class variables: variable names of inputs, parameters and outputs
    inputs_names = ["load"]
    outputs_names = ["rpm"]

    def __init__(self, L_min, omega_max):
        """Initializes the simulator."""
        super(ModelA1, self).__init__()
        self.L_min = L_min
        self.omega_max = omega_max

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        return {"L_min": self.L_min, "omega_max": self.omega_max}

    def _simulate(self, load, L_min, omega_max):
        """
        Simulate the steady-state fan speed using Model A1.

        This function computes the steady-state angular speed of the fan in radians per second
        using Model A1 and then converts the result to revolutions per minute (rpm).

        Parameters
        ----------
        load : float or array-like
            The fan load (or time series of loads) applied to the fan.
        L_min : float
            The minimum effective load. For load < L_min, L_min is used.
        omega_max : float
            The maximum angular speed of the fan (in rpm).

        Returns
        -------
        float or array-like
            The simulated fan speed in rpm.

        Notes
        -----
        Conversion from rad/s to rpm is performed using the factor (30 / π).
        This method is invoked by the class method `simulate_from_inputs`.
        """
        rads = model_a1(L=load, L_min=L_min, omega_max=omega_max * np.pi / 30)
        # Transform from rad/s to rpm
        return rads / np.pi * 30


class ModelA2(Simulator):
    """Simulator of the fan speed dynamics given the load through time.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model A2 in Appendix IV.1 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    """

    # Class variables: variable names of inputs, parameters and outputs
    inputs_names = ["load", "timestamp"]
    outputs_names = ["rpm"]

    def __init__(
        self,
        # Parameters
        I,
        tau,
        K,
        # Parameters for the ODE solver
        omega_0,
        simulation_steps=100,
    ):
        """Initializes the simulator."""
        super(ModelA2, self).__init__()
        # Parameters
        self.I = I
        self.tau = tau
        self.K = K
        self.omega_0 = omega_0
        self.simulation_steps = simulation_steps

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        return {
            "I": self.I,
            "tau": self.tau,
            "K": self.K,
            "omega_0": self.omega_0,
            "simulation_steps": self.simulation_steps,
        }

    def _simulate(self, load, timestamp, I, tau, K, omega_0, simulation_steps):
        """
        Simulate the dynamic behavior of the fan using Model A2.

        This function integrates the fan’s speed over time using the torque-balance differential
        equation from Model A2. An ODE solver (Euler's method) is used with a specified number of
        simulation steps, and the output is converted from rad/s to rpm.

        See Model A2 in Appendix IV.1 for more details (https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332).

        Parameters
        ----------
        load : array-like
            Time series of the fan load.
        timestamp : array-like
            The time points (in seconds) at which the simulation is computed.
        I : float
            The moment of inertia of the fan (kg·m²).
        tau : function
            A function that computes the motor torque as a function of load.
        K : float
            The drag constant representing the drag of the fan.
        omega_0 : float
            The initial angular speed of the fan (in rpm; note that the internal model works in rad/s).
        simulation_steps : int
            The number of steps used by the ODE solver (Euler's method).

        Returns
        -------
        float or array-like
            The simulated fan speed in rpm.

        Notes
        -----
        Conversion from rad/s to rpm is performed using the factor (30 / π).
        This method is invoked by the class method `simulate_from_inputs`.

        The torque function (tau) used in the original paper (see model A2 in Appendix IV) is given by:

        ```python
        def tau(load, C_min, C_max, L_min, T):
            load = np.atleast_1d(load)
            torques = T * (C_min + np.maximum(L_min, load) ** 3 * (C_max - C_min) - C_min)
            torques[load == 0] = 0
            return torques if len(load) > 1 else torques[0]
        ```
        """
        rads = model_a2(
            loads=load,
            timestamps=timestamp,
            I=I,
            tau=tau,
            K=K,
            omega_0=omega_0 * np.pi / 30,
            simulation_steps=simulation_steps,
        )
        # Transform from rad/s to rpm
        return rads / np.pi * 30


class ModelB1(Simulator):
    """Simulator of the fan current given its load.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model B1 in Appendix IV.1 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    """

    # Class variables: variable names of inputs, parameters and outputs
    inputs_names = ["load"]
    outputs_names = ["current"]

    def __init__(
        self,
        # Parameters
        C_min,
        C_max,
        L_min,
    ):
        """Initializes the simulator."""
        super(ModelB1, self).__init__()
        # Parameters
        self.C_min = C_min
        self.C_max = C_max
        self.L_min = L_min

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        return {
            "C_min": self.C_min,
            "C_max": self.C_max,
            "L_min": self.L_min,
        }

    def _simulate(self, load, C_min, C_max, L_min):
        """Simulate the electrical current drawn by the fan using Model B1.

        This function computes the current drawn by the fan based on
        the applied load using the cubic relationship described in
        Model B1. When the load is zero, the no-load current is
        returned.

        Parameters
        ----------
        load : float or array-like
            The fan load.
        C_min : float
            The no-load current of the fan.
        C_max : float
            The maximum current drawn by the fan at full load.
        L_min : float
            The minimum effective load; for loads below this threshold, L_min is used.

        Returns
        -------
        float or array-like
            The simulated current (in Amperes) drawn by the fan.

        Notes
        -----
        This method is invoked by the class method `simulate_from_inputs`.
        """
        load = np.atleast_1d(load)
        current = C_min + np.maximum(load, L_min) ** 3 * (C_max - C_min)
        current[load == 0] = C_min
        return current if len(load) > 1 else current[0]


class SimA1C2(Simulator):
    """Simulator for the pressure measurements from the downwind barometer
    in the wind tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Models A1 and C2 in the paper
    "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    """

    # Class variables: variable names of inputs and outputs
    inputs_names = ["load_in", "load_out", "pressure_ambient"]
    outputs_names = ["pressure_downwind", "rpm_in", "rpm_out"]

    def __init__(
        self,
        # Parameters for model A1
        L_min,
        omega_max,
        # Parameters for model C2
        S_max,
        Q_max,
        r,
        # Sensor noise
        barometer_error,  # The barometer offset
        barometer_precision,  # The std. of the barometer sensor noise
        random_state=42,
    ):
        """Initializes the simulator."""
        super(SimA1C2, self).__init__()

        # Parameters for model A1
        self.L_min = L_min
        self.omega_max = omega_max

        # Parameters for model C2
        self.S_max = S_max
        self.Q_max = Q_max
        self.r = r

        # Sensor noise
        self.barometer_error = barometer_error
        self.barometer_precision = barometer_precision
        self.random_state = random_state

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        return {
            "L_min": self.L_min,
            "omega_max": self.omega_max,
            "S_max": self.S_max,
            "Q_max": self.Q_max,
            "r": self.r,
            "barometer_error": self.barometer_error,
            "barometer_precision": self.barometer_precision,
            "random_state": self.random_state,
        }

    def _simulate(
        self,
        load_in,
        load_out,
        pressure_ambient,
        L_min,
        omega_max,
        # Parameters for model C2
        S_max,
        Q_max,
        r,
        # Sensor noise
        barometer_error,  # The barometer offset
        barometer_precision,  # The std. of the barometer sensor noise
        random_state,
    ):
        """
        Simulate wind tunnel conditions using Models A1 and C2.

        This function simulates the static pressure inside a wind tunnel by combining the
        steady-state fan speed model (Model A1) with the static pressure model (Model C2). It computes
        the downwind pressure and the speeds of both the intake and exhaust fans, converting the speeds
        from rad/s to rpm.

        Parameters
        ----------
        load_in : float or array-like
            The load applied to the intake fan.
        load_out : float or array-like
            The load applied to the exhaust fan.
        pressure_ambient : float
            The ambient static pressure (Pa) outside the wind tunnel.
        L_min : float
            The minimum effective load used in Model A1.
        omega_max : float
            The maximum fan speed (in rpm).
        S_max : float
            The maximum static pressure produced by the fan at full speed.
        Q_max : float
            The maximum airflow produced by the fan (in m³/s).
        r : float
            The airflow ratio parameter used in Model C2.
        barometer_error : float
            The offset error of the barometer sensor.
        barometer_precision : float
            The standard deviation of the barometer sensor noise.
        random_state : int or RandomState
            Seed or random state for simulating sensor noise.

        Returns
        -------
        tuple
            A tuple containing:
                - pressure_downwind (float or array-like): Simulated downwind static pressure (Pa).
                - rpm_in (float or array-like): Intake fan speed in rpm.
                - rpm_out (float or array-like): Exhaust fan speed in rpm.

        Notes
        -----
        Fan speeds are converted from rad/s to rpm using the factor (30 / π).
        This method is invoked by the class method `simulate_from_inputs`.
        """
        pressure_downwind, omega_in, omega_out = simulator_a1_c2(
            load_in=load_in,
            load_out=load_out,
            hatch=None,  # ignore hatch input
            P_amb=pressure_ambient,
            L_min=L_min,
            omega_max=omega_max * np.pi / 30,  # convert to rad/s
            # Parameters for model C2
            S_max=S_max,
            Q_max=Q_max,
            r=r,
            # Sensor noise
            barometer_error=barometer_error,  # The barometer offset
            barometer_precision=barometer_precision,  # The std. of the barometer sensor noise
            random_state=random_state,
        )
        rpm_in = omega_in / np.pi * 30
        rpm_out = omega_out / np.pi * 30
        return pressure_downwind, rpm_in, rpm_out


class SimA1C3(Simulator):
    """Simulator for the pressure measurements from the downwind barometer
    in the wind tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Models A1 and C3 in the paper
    "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    See more details in the `_simulate` method.

    """

    # Class variables: variable names of inputs, parameters, and outputs
    inputs_names = ["load_in", "load_out", "hatch", "pressure_ambient"]
    outputs_names = ["pressure_downwind", "rpm_in", "rpm_out"]

    def __init__(
        self,
        # Parameters for model A1
        L_min,
        omega_max,
        # Parameters for model C3
        S_max,
        Q_max,
        r_0,
        beta,
        # Sensor noise
        barometer_error,  # The barometer offset
        barometer_precision,  # The std. of the barometer sensor noise
        random_state=42,
    ):
        """Initializes the simulator with the given parameters. . See the
        docstring for `_simulate` for a full description of the simulator
        parameters."""
        super(SimA1C3, self).__init__()

        # Parameters for models A1 and C3
        self.L_min = L_min
        self.omega_max = omega_max
        self.S_max = S_max
        self.Q_max = Q_max
        self.r_0 = r_0
        self.beta = beta

        # Sensor noise parameters
        self.barometer_error = barometer_error
        self.barometer_precision = barometer_precision
        self.random_state = random_state

    def parameters(self):
        """Return a dictionary with the simulator parameters and their values."""
        return {
            "L_min": self.L_min,
            "omega_max": self.omega_max,
            "S_max": self.S_max,
            "Q_max": self.Q_max,
            "r_0": self.r_0,
            "beta": self.beta,
            "barometer_error": self.barometer_error,
            "barometer_precision": self.barometer_precision,
            "random_state": self.random_state,
        }

    def _simulate(
        self,
        load_in,
        load_out,
        hatch,
        pressure_ambient,
        L_min,
        omega_max,
        S_max,
        Q_max,
        r_0,
        beta,
        barometer_error,
        barometer_precision,
        random_state,
    ):
        """
        Simulate wind tunnel conditions using Models A1 and C3.

        This function computes the downwind static pressure as well as the intake and exhaust fan speeds
        by combining the steady-state fan model (Model A1) with the static pressure model that accounts for
        the hatch position (Model C3). The fan speeds are converted from rad/s to rpm.

        Parameters
        ----------
        load_in : float or array-like
            The load applied to the intake fan.
        load_out : float or array-like
            The load applied to the exhaust fan.
        hatch : float or array-like
            The hatch position, in degrees.
        pressure_ambient : float
            The ambient static pressure (Pa) outside the wind tunnel.
        L_min : float
            The minimum effective load parameter for the fan model.
        omega_max : float
            The maximum fan speed (in rpm).
        S_max : float
            The maximum static pressure produced by the fan at full speed.
        Q_max : float
            The maximum airflow produced by the fan (in m³/s).
        r_0 : float
            The baseline airflow ratio when the hatch is closed.
        beta : float
            The coefficient representing the linear effect of the hatch position on the airflow ratio.
        barometer_error : float
            The offset error of the barometer sensor.
        barometer_precision : float
            The standard deviation of the barometer sensor noise.
        random_state : int or RandomState
            Seed or random state for simulating sensor noise.

        Returns
        -------
        tuple
            A tuple containing:
                - pressure_downwind (float or array-like): Simulated downwind static pressure (Pa).
                - rpm_in (float or array-like): Intake fan speed in rpm.
                - rpm_out (float or array-like): Exhaust fan speed in rpm.

        Notes
        -----
        Fan speeds are converted from rad/s to rpm using the factor (30 / π).
        This method is invoked by the class method `simulate_from_inputs`.

        """
        pressure_downwind, omega_in, omega_out = simulator_a1_c3(
            load_in=load_in,
            load_out=load_out,
            hatch=hatch,
            P_amb=pressure_ambient,
            L_min=L_min,
            omega_max=omega_max * np.pi / 30,  # convert to rad/s
            S_max=S_max,
            Q_max=Q_max,
            r_0=r_0,
            beta=beta,
            barometer_error=barometer_error,
            barometer_precision=barometer_precision,
            random_state=random_state,
        )
        rpm_in = omega_in / np.pi * 30
        rpm_out = omega_out / np.pi * 30
        return pressure_downwind, rpm_in, rpm_out


class SimA2C3(Simulator):
    """Simulator for the pressure measurements from the downwind barometer
    in the wind tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Models A2 and C3 in the paper
    "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332

    """

    # Class variables: variable names of inputs, parameters, and outputs
    inputs_names = ["load_in", "load_out", "hatch", "pressure_ambient", "timestamp"]
    outputs_names = ["pressure_downwind", "omega_in", "omega_out"]

    def __init__(
        self,
        # Parameters for model A2
        I,
        tau,
        K,
        omega_in_0,
        omega_out_0,
        # Parameters for model C3
        S_max,
        omega_max,
        Q_max,
        r_0,
        beta,
        # Sensor noise
        barometer_error,  # The barometer offset
        barometer_precision,  # The std. of the barometer sensor noise
        random_state=42,
        # For the ODE solver of model A2
        simulation_steps=100,
    ):
        """Initializes the simulator with the given parameters. See the
        docstring for `_simulate` for a full description of the simulator
        parameters."""
        super(SimA2C3, self).__init__()

        # Parameters for models A2 and C3
        self.I = I
        self.tau = tau
        self.K = K
        self.omega_in_0 = omega_in_0
        self.omega_out_0 = omega_out_0
        self.S_max = S_max
        self.omega_max = omega_max
        self.Q_max = Q_max
        self.r_0 = r_0
        self.beta = beta

        # Sensor noise parameters
        self.barometer_error = barometer_error
        self.barometer_precision = barometer_precision
        self.random_state = random_state

        # ODE solver parameter
        self.simulation_steps = simulation_steps

    def parameters(self):
        """Return a dictionary with the simulator parameters and their values."""
        return {
            "I": self.I,
            "tau": self.tau,
            "K": self.K,
            "omega_in_0": self.omega_in_0,
            "omega_out_0": self.omega_out_0,
            "S_max": self.S_max,
            "omega_max": self.omega_max,
            "Q_max": self.Q_max,
            "r_0": self.r_0,
            "beta": self.beta,
            "barometer_error": self.barometer_error,
            "barometer_precision": self.barometer_precision,
            "random_state": self.random_state,
            "simulation_steps": self.simulation_steps,
        }

    def _simulate(
        self,
        # Inputs
        load_in,
        load_out,
        hatch,
        pressure_ambient,
        timestamp,
        # Parameters for model A2
        I,
        tau,
        K,
        omega_in_0,
        omega_out_0,
        # Parameters for model C3
        S_max,
        omega_max,
        Q_max,
        r_0,
        beta,
        # Parameters for noise simulation
        barometer_error,
        barometer_precision,
        random_state,
        simulation_steps,
    ):
        """
        Simulate dynamic wind tunnel behavior using Models A2 and C3.

        This function combines the dynamic fan speed simulation from Model A2 with the static
        pressure model that accounts for hatch position (Model C3). It computes the evolution
        of the downwind static pressure along with the intake and exhaust fan speeds over time.
        The fan speeds are internally calculated in radians per second and then converted to
        revolutions per minute (rpm) using the conversion factor (30 / π).

        The inputs (load_in, load_out, hatch, pressure_ambient and timestamp) are time-series.

        Parameters
        ----------
        load_in : float or array-like
            The load applied to the intake fan.
        load_out : float or array-like
            The load applied to the exhaust fan.
        hatch : float or array-like
            The hatch position affecting the system impedance.
        pressure_ambient : float
            The ambient static pressure (Pa) outside the wind tunnel.
        timestamp : array-like
            Time stamps (in seconds) at which the simulation is evaluated.
        I : float
            The moment of inertia of the fan (kg·m²), used in Model A2.
        tau : function
            A function that computes the motor torque as a function of load for Model A2.
        K : float
            The drag constant in the torque-balance equation of Model A2.
        omega_in_0 : float
            The initial intake fan speed (in rpm; note that internal computations use rad/s).
        omega_out_0 : float
            The initial exhaust fan speed (in rpm; note that internal computations use rad/s).
        S_max : float
            The maximum static pressure produced by the fan at full speed (Pa) in Model C3.
        omega_max : float
            The maximum fan speed (in rpm) in Model C3.
        Q_max : float
            The maximum airflow produced by the fan (m³/s) in Model C3.
        r_0 : float
            The baseline airflow ratio when the hatch is closed in Model C3.
        beta : float
            The coefficient representing the linear effect of the hatch position on the airflow ratio.
        barometer_error : float
            The offset error of the barometer sensor.
        barometer_precision : float
            The standard deviation of the barometer sensor noise.
        random_state : int or RandomState
            Seed or random state for simulating sensor noise.
        simulation_steps : int
            The number of simulation steps used by the internal ODE solver (Euler's method).

        Returns
        -------
        tuple
            A tuple containing:
                - pressure_downwind (float or array-like): The simulated downwind static pressure (Pa).
                - rpm_in (float or array-like): The simulated intake fan speed in rpm.
                - rpm_out (float or array-like): The simulated exhaust fan speed in rpm.

        Notes
        -----
        Fan speeds are converted from rad/s to rpm using the factor (30 / π).
        This method is invoked by the class method `simulate_from_inputs`.
        """
        pressure_downwind, omega_in, omega_out = simulator_a2_c3(
            load_in=load_in,
            load_out=load_out,
            hatch=hatch,
            P_amb=pressure_ambient,
            timestamps=timestamp,
            I=I,
            tau=tau,
            C=K,
            omega_in_0=omega_in_0 * np.pi / 30,  # convert to rad/s
            omega_out_0=omega_out_0 * np.pi / 30,
            S_max=S_max,
            omega_max=omega_max * np.pi / 30,
            Q_max=Q_max,
            r_0=r_0,
            beta=beta,
            barometer_error=barometer_error,
            barometer_precision=barometer_precision,
            random_state=random_state,
            simulation_steps=simulation_steps,
        )
        rpm_in = omega_in / np.pi * 30
        rpm_out = omega_out / np.pi * 30
        return pressure_downwind, rpm_in, rpm_out


# --------------------------------------------------------------------
# Mechanistic models of the wind tunnel processes

# These are called from the _simulate function of each simulator
# class. This redundancy is done to clearly separate the unit conversion
# between RPMs and rad/s for models A1 and A2.

# For the derivation and more details about each model and its
# parameters, see Appendix IV.1 of the paper "Causal chambers as a
# real-world physical testbed for AI Methodology" (2025) by Gamella et
# al.

# Link for direct access:
# https://arxiv.org/pdf/2404.11341#page=28&zoom=100,57,332


def model_a1(
    # Input
    L,
    # Parameters
    L_min,
    omega_max,
):
    """Model A1 of the steady-state fan speed given the load."""
    L = np.atleast_1d(L)
    omega = np.maximum(L, L_min) * omega_max
    omega[L == 0] = 0
    return omega if len(L) > 1 else omega[0]


def model_a2(
    # Input
    loads,
    # Parameters
    I,
    tau,
    K,
    # Parameters for the ODE solver
    omega_0,
    timestamps,
    simulation_steps,
):
    """Model A2 of the fan-speed dynamics given a time-series of the fan
    load."""
    # Compute torque at each time point
    torques = tau(loads)

    # Compute speed at each time point by solving the ODE
    omegas = np.zeros_like(loads, dtype=float)
    omegas[0] = omega_0
    for i in range(1, len(loads)):
        timestep = timestamps[i] - timestamps[i - 1]
        torque = torques[i]
        omega = omegas[i - 1]
        dt = timestep / simulation_steps
        for _ in range(simulation_steps):
            d_omega = 1 / I * (torque - K * omega**2)
            omega += dt * d_omega
        omegas[i] = omega
    return omegas


def model_c1(
    # Input
    omega_in,
    omega_out,
    P_amb,
    # Parameters
    S_max,
    omega_max,
):
    """Model C1 of the effect of the loads and hatch on the downwind barometer"""
    return (
        P_amb
        + S_max * (omega_in / omega_max) ** 2
        - S_max * (omega_out / omega_max) ** 2
    )


def model_c2(
    # Input
    omega_in,
    omega_out,
    P_amb,
    # Parameters
    S_max,
    omega_max,
    Q_max,
    r,
):
    """Model C2 of the effect of the loads and hatch on the downwind barometer"""
    S_r = lambda omega: _S_r(omega, r, Q_max, S_max, omega_max)
    return P_amb + S_r(omega_in) - S_r(omega_out)


def model_c3(
    # Input
    omega_in,
    omega_out,
    H,
    P_amb,
    # Parameters
    S_max,
    omega_max,
    Q_max,
    r_0,
    beta,
):
    """Model C3 of the effect of the loads and hatch on the downwind barometer"""
    # Compute impedance (airflow ratio) as a function of the hatch position H
    r = np.minimum(1, r_0 + beta * H / 45)
    S_rh = lambda omega: _S_r(omega, r, Q_max, S_max, omega_max)
    return P_amb + S_rh(omega_in) - S_rh(omega_out)


def _S_r(
    omega,
    r,
    Q_max,
    S_max,
    omega_max,
):
    """Used in models C2 and C3 to compute the static pressure produced by
    the fan, as the intersection of the impedance curve (for airflow
    ratio r) and the (linear) PQ-characteristic. See the manuscript for more details.

    """
    Z = S_max / Q_max**2 * (1 - r) / r**2
    # Find the intersection of the impedance curve S = ZQ^2 and the PQ-characteristic
    #   solve aQ^2 + bQ + c = 0 using the quadratic formula, where
    a = Z
    b = (omega / omega_max) * S_max / Q_max
    c = -((omega / omega_max) ** 2) * S_max
    # Intersection (Q,S)
    Q = (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a  # produced airflow
    S = Z * Q**2  # produced static pressure
    return S


def simulator_a1_c2(
    # Input
    load_in,
    load_out,
    hatch,
    P_amb,
    # Parameters for model A1
    L_min,
    omega_max,
    # Parameters for model C2
    S_max,
    Q_max,
    r,
    # Sensor noise
    barometer_error,  # The barometer offset
    barometer_precision,  # The std. of the barometer sensor noise
    random_state=42,
):
    load_in, load_out = np.array(load_in), np.array(load_out)
    omega_in = model_a1(load_in, L_min, omega_max)
    omega_out = model_a1(load_out, L_min, omega_max)
    P_dw = model_c2(omega_in, omega_out, P_amb, S_max, omega_max, Q_max, r)
    rng = np.random.default_rng(random_state)
    P_dw += rng.normal(barometer_error, barometer_precision, size=len(P_dw))
    return P_dw, omega_in, omega_out


def simulator_a1_c3(
    # Input
    load_in,
    load_out,
    hatch,
    P_amb,
    # Parameters for model A1
    L_min,
    omega_max,
    # Parameters for model C3
    S_max,
    Q_max,
    r_0,
    beta,
    # Sensor noise
    barometer_error,  # The barometer offset
    barometer_precision,  # The std. of the barometer sensor noise
    random_state=42,
):
    load_in, load_out = np.array(load_in), np.array(load_out)
    omega_in = model_a1(load_in, L_min, omega_max)
    omega_out = model_a1(load_out, L_min, omega_max)
    P_dw = model_c3(
        omega_in, omega_out, hatch, P_amb, S_max, omega_max, Q_max, r_0, beta
    )
    rng = np.random.default_rng(random_state)
    P_dw += rng.normal(barometer_error, barometer_precision, size=len(P_dw))
    return P_dw, omega_in, omega_out


def simulator_a2_c3(
    # Input
    load_in,
    load_out,
    hatch,
    P_amb,
    # Parameters for model A1
    I,
    tau,
    C,
    timestamps,
    omega_in_0,
    omega_out_0,
    # Parameters for model C3
    S_max,
    omega_max,
    Q_max,
    r_0,
    beta,
    # Sensor noise
    barometer_error,  # The barometer offset
    barometer_precision,  # The std. of the barometer sensor noise
    random_state=42,
    # For the ODE solver of model a2
    simulation_steps=100,
):
    load_in, load_out = np.array(load_in), np.array(load_out)
    omega_in = model_a2(load_in, I, tau, C, omega_in_0, timestamps, simulation_steps)
    omega_out = model_a2(load_out, I, tau, C, omega_out_0, timestamps, simulation_steps)
    P_dw = model_c3(
        omega_in, omega_out, hatch, P_amb, S_max, omega_max, Q_max, r_0, beta
    )
    rng = np.random.default_rng(random_state)
    P_dw += rng.normal(barometer_error, barometer_precision, size=len(P_dw))
    return P_dw, omega_in, omega_out



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Numpy implementation of the deterministic simulator of the
light-tunnel sensors.
"""

import numpy as np
from copy import deepcopy
from causalchamber.simulators import Simulator


class Deterministic(Simulator):
    # Class variables: variable names of inputs, parameters and outputs
    inputs_names = [
        "red",
        "green",
        "blue",
        "pol_1",
        "pol_2",
        "diode_ir_1",
        "diode_ir_2",
        "diode_ir_3",
        "diode_vis_1",
        "diode_vis_2",
        "diode_vis_3",
        "t_ir_1",
        "t_ir_2",
        "t_ir_3",
        "t_vis_1",
        "t_vis_2",
        "t_vis_3",
        "v_c",
        "v_angle_1",
        "v_angle_2",
    ]
    outputs_names = [
        "ir_1",
        "vis_1",
        "ir_2",
        "vis_2",
        "ir_3",
        "vis_3",
        "current",
        "angle_1",
        "angle_2",
    ]
    parameters_names = [
        "S",
        "d1",
        "d2",
        "d3",
        "Ts",
        "Tp",
        "Tc",
        "Q",
        "C0",
        "A",
        "a1",
        "a2",
    ]

    def __init__(self, S, d1, d2, d3, Ts, Tp, Tc, Q, C0, A, a1, a2):
        super(Deterministic, self).__init__()
        self.S = deepcopy(S)
        self.d1 = d1
        self.d2 = d2
        self.d3 = d3
        self.Ts = deepcopy(Ts)
        self.Tp = deepcopy(Tp)
        self.Tc = deepcopy(Tc)
        self.Q = deepcopy(Q)
        self.C0 = deepcopy(C0)
        self.A = deepcopy(A)
        self.a1 = deepcopy(a1)
        self.a2 = deepcopy(a2)

    def parameters(self):
        return {
            "S": self.S,
            "d1": self.d1,
            "d2": self.d2,
            "d3": self.d3,
            "Ts": self.Ts,
            "Tp": self.Tp,
            "Tc": self.Tc,
            "Q": self.Q,
            "C0": self.C0,
            "A": self.A,
            "a1": self.a1,
            "a2": self.a2,
        }

    def _simulate(
        self,
        # Simulator inputs
        red,
        green,
        blue,
        pol_1,
        pol_2,
        diode_ir_1,
        diode_ir_2,
        diode_ir_3,
        diode_vis_1,
        diode_vis_2,
        diode_vis_3,
        t_ir_1,
        t_ir_2,
        t_ir_3,
        t_vis_1,
        t_vis_2,
        t_vis_3,
        v_c,
        v_angle_1,
        v_angle_2,
        # Simulator parameters
        S,
        d1,
        d2,
        d3,
        Ts,
        Tp,
        Tc,
        Q,
        C0,
        A,
        a1,
        a2,
    ):
        """Simulates the light-intensity measurements from the sensors in the
        light tunnel. For the complete derivation and details of all
        parameters, see Appendix C of the paper "Sanity Checking
        Causal Representation Learning on a Simple Real-World System"
        (2025) by Juan L. Gamella*, Simon Bing*, and Jakob Runge.

        Link for direct access: TODO

        The function takes as input the values of the tunnel actuators
        (light-source color, polarizer positions and light-sensor
        parameters), various simulation parameters, and returns the
        uncalibrated infrared and visible-light intensity
        measurements, as well as the current and angle measurements.

        Parameters
        ----------
        red, green, blue : array_like
            1D arrays representing the brightness settings of the red,
            green, and blue LEDs respectively, each in the range [0,
            255].
        pol_1, pol_2 : array_like
            1D arrays representing the positions of the tunnel's
            polarizers in degrees, in the range [-180, 180].
        diode_ir_1, diode_ir_2, diode_ir_3 : array_like
            1D arrays representing the sizes of the infrared
            photodiodes used by the light sensors. Values are 0
            (small), 1 (medium), or 2 (large).
        diode_vis_1, diode_vis_2, diode_vis_3 : array_like
            1D arrays representing the sizes of the visible-light
            photodiodes used by the light sensors. Values are 0
            (small), 1 (medium), or 2 (large).
        t_ir_1, t_ir_2, t_ir_3 : array_like
            1D arrays representing the exposure time for the infrared
            sensors during measurements. Values are integers in the
            range [0, 3].
        t_vis_1, t_vis_2, t_vis_3 : array_like
            1D arrays representing the exposure time for the
            visible-light sensors during measurements. Values are
            integers in the range [0, 3].
        S : array_like, shape (2, 3)
            The linear response matrix of the photodiodes to the red,
            green, and blue light.  The first row corresponds to the
            response of the smallest IR photodiode, and the second row
            corresponds to the smallest visible-light photodiode.
        d1, d2, d3 : float
            The distances (in millimeters) from the tunnel's light
            source to the first, second, and third sensors
            respectively.
        Ts : np.ndarray
            The transmission rate (scaling of light intensity) for
            each color wavelength of the first polarizer (vector of
            length 3).
        Tp : np.ndarray
            Transmission rates for each color when the polarizers are
            aligned (vector of length 3).
        Tc : np.ndarray
            Transmission rates for each color when the polarizers are
            perpendicular(vector of length 3).
        Q : np.ndarray
            1 x 3 matrix that encodes the linear response of the
            current sensor to the brightness setting of each color.
        C0 : float
            The current drawn by the light source when it is turned
            off, i.e., `red=green=blue=0`. The value corresponds to
            the uncalibrated measurement when `v_c=5`.
        A : float
            The linear coefficient relating the change in polarizer
            position to the change in the voltage reading of the angle
            sensor, when the reference voltages are `v_angle_1=5` and
            `v_angle_2=5`, respectively.
        a1, a2: float
            The zero point of the angle sensors, i.e., the voltage
            reading when the polarizer positions are `pol_1=0` and
            `pol_2=0`, respectively, and the reference voltages are
            `v_angle_1=v_angle_2=5`.

        Returns
        -------
        ir_1 : numpy.ndarray
            Uncalibrated infrared light-intensity measurements from the first sensor.
        vis_1 : numpy.ndarray
            Uncalibrated visible light-intensity measurements from the first sensor.
        ir_2 : numpy.ndarray
            Uncalibrated infrared light-intensity measurements from the second sensor.
        vis_2 : numpy.ndarray
            Uncalibrated visible light-intensity measurements from the second sensor.
        ir_3 : numpy.ndarray
            Uncalibrated infrared light-intensity measurements from the third sensor.
        vis_3 : numpy.ndarray
            Uncalibrated visible light-intensity measurements from the third sensor.
        current : numpy.ndarray
            Simulated measurements of the current drawn by the light source, in Amperes.
        angle_1 : numpy.ndarray
            Simulated measurements of the angle sensor for the first polarizer.
        angle_2 : numpy.ndarray
            Simulated measurements of the angle sensor for the second polarizer.

        """

        RGB = np.array([red, green, blue])

        # First sensor
        SRGB = S @ RGB
        ir_1 = 2 ** (diode_ir_1 + t_ir_1) * SRGB[0]
        vis_1 = 2 ** (diode_vis_1 + t_vis_1) * SRGB[1]

        # Second sensor
        SRGB = S @ np.diag(Ts) @ ((d1 / d2) ** 2 * RGB)
        ir_2 = 2 ** (diode_ir_2 + t_ir_2) * SRGB[0]
        vis_2 = 2 ** (diode_vis_2 + t_vis_2) * SRGB[1]

        # Third sensor
        malus = np.atleast_2d(np.cos(np.deg2rad(pol_1 - pol_2)) ** 2)
        CRGB = (d1 / d3) ** 2 * RGB
        CRGB = np.diag(Tp - Tc) @ (malus * CRGB) + np.diag(Tc) @ CRGB
        SRGB = S @ CRGB
        ir_3 = 2 ** (diode_ir_3 + t_ir_3) * SRGB[0]
        vis_3 = 2 ** (diode_vis_3 + t_vis_3) * SRGB[1]

        # Current
        current = (Q @ RGB + C0) * 5.0 / v_c

        # Angles
        angle_1 = np.minimum((A * pol_1 + a1) * 5.0 / v_angle_1, 1023)
        angle_2 = np.minimum((A * pol_2 + a2) * 5.0 / v_angle_2, 1023)

        return ir_1, vis_1, ir_2, vis_2, ir_3, vis_3, current, angle_1, angle_2



# MIT License

# Copyright (c) 2025 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from causalchamber.simulators import Simulator
import numpy as np

# --------------------------------------------------------------
# Simulators


class ModelF1(Simulator):
    """Simulator of the images produced by the light tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model F1 in Appendix IV.2.2 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=32&zoom=100,57,670

    """

    inputs_names = [
        "red",
        "green",
        "blue",
        "pol_1",
        "pol_2",
    ]
    outputs_names = ["image"]

    def __init__(
        self,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """
        Initialize the simulator by storing its parameters.
        """
        super(ModelF1, self).__init__()
        # Store the simulator's parameters
        self.center_x = center_x
        self.center_y = center_y
        self.radius = radius
        self.offset = offset
        self.image_size = image_size

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        params = {
            "center_x": self.center_x,
            "center_y": self.center_y,
            "radius": self.radius,
            "offset": self.offset,
            "image_size": self.image_size,
        }
        return params

    def _simulate(
        self,
        red,
        green,
        blue,
        pol_1,
        pol_2,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """Simulates a synthetic image using model_f1, generating a colored
        hexagon over a black background.

        Parameters
        ----------
        red : np.ndarray
            Brightness of the red LEDs of the light source.
        green : np.ndarray
            Brightness of the green LEDs of the light source.
        blue : np.ndarray
            Brightness of the blue LEDs of the light source.
        pol_1 : np.ndarray
            Angle of the first polarizer.
        pol_2 : np.ndarray
            Angle of the second polarizer.
        center_x : float
            X-coordinate of the hexagon's center in the image.
        center_y : float
            Y-coordinate of the hexagon's center in the image.
        radius : float
            Radius of the circumference encompassing the hexagon.
        offset : float
            Rotation of the hexagon in degrees.
        image_size : int
            Size of the synthetic image in pixels (i.e., image_size x image_size pixels).

        Returns
        -------
        np.ndarray
            Generated synthetic images with dimensions [n_images,
            image_size, image_size, 3], where n_images is the length of
            the inputs (i.e., red, green, blue, pol_1, pol_2).

        """
        N = len(red)
        images = np.ones((N, image_size, image_size, 3))
        pol_1, pol_2 = np.deg2rad(pol_1), np.deg2rad(pol_2)
        # Color
        malus_factor = np.cos(pol_1 - pol_2) ** 2
        red = red / 255 * malus_factor
        green = green / 255 * malus_factor
        blue = blue / 255 * malus_factor
        images[:, :, :, 0] *= red[:, np.newaxis, np.newaxis]
        images[:, :, :, 1] *= green[:, np.newaxis, np.newaxis]
        images[:, :, :, 2] *= blue[:, np.newaxis, np.newaxis]

        # Apply hexagon mask
        mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
        images *= mask[np.newaxis, :, :, np.newaxis]
        return clip(images)


class ModelF2(Simulator):
    """
    Simulator of the images produced by the light tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model F2 in Appendix IV.2.2 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=32&zoom=100,57,670
    """

    inputs_names = [
        "red",
        "green",
        "blue",
        "pol_1",
        "pol_2",
    ]
    outputs_names = ["image"]

    def __init__(
        self,
        S,
        w_r,
        w_g,
        w_b,
        exposure,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """
        Initialize the simulator by storing its parameters.
        """
        super(ModelF2, self).__init__()
        # Store the simulator's parameters
        self.S = S
        self.w_r = w_r
        self.w_g = w_g
        self.w_b = w_b
        self.exposure = exposure
        self.center_x = center_x
        self.center_y = center_y
        self.radius = radius
        self.offset = offset
        self.image_size = image_size

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        params = {
            "S": self.S,
            "w_r": self.w_r,
            "w_g": self.w_g,
            "w_b": self.w_b,
            "exposure": self.exposure,
            "center_x": self.center_x,
            "center_y": self.center_y,
            "radius": self.radius,
            "offset": self.offset,
            "image_size": self.image_size,
        }
        return params

    def _simulate(
        self,
        red,
        green,
        blue,
        pol_1,
        pol_2,
        S,
        w_r,
        w_g,
        w_b,
        exposure,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """Simulates a synthetic image using model_f2, generating a colored
        hexagon over a black background.

        Parameters
        ----------
        red : np.ndarray
            Brightness of the red LEDs of the light source.
        green : np.ndarray
            Brightness of the green LEDs of the light source.
        blue : np.ndarray
            Brightness of the blue LEDs of the light source.
        pol_1 : np.ndarray
            Angle of the first polarizer.
        pol_2 : np.ndarray
            Angle of the second polarizer.
        S : np.ndarray
            3x3 matrix modeling the sensor response to light of each color.
        w_r : float
            White-balance correction factor for the red channel.
        w_g : float
            White-balance correction factor for the green channel.
        w_b : float
            White-balance correction factor for the blue channel.
        exposure : float
            Exposure parameter affected by camera settings (aperture, shutter speed, ISO).
        center_x : float
            X-coordinate of the hexagon's center in the image.
        center_y : float
            Y-coordinate of the hexagon's center in the image.
        radius : float
            Radius of the circumference encompassing the hexagon.
        offset : float
            Rotation of the hexagon in degrees.
        image_size : int
            Size of the synthetic image in pixels (i.e., image_size x image_size pixels).

        Returns
        -------
        np.ndarray
            Generated synthetic images with dimensions [n_images,
            image_size, image_size, 3], where n_images is the length of
            the inputs (i.e., red, green, blue, pol_1, pol_2).

        """
        N = len(red)
        images = np.ones((N, image_size, image_size, 3))
        # Transform parameters
        pol_1, pol_2 = np.deg2rad(pol_1), np.deg2rad(pol_2)
        r = np.atleast_1d(red / 255)
        g = np.atleast_1d(green / 255)
        b = np.atleast_1d(blue / 255)

        # Ccalculate color
        malus_factor = np.cos(pol_1 - pol_2) ** 2
        W = np.diag([w_r, w_g, w_b])
        color = exposure * W @ S @ np.array([r, g, b]) * malus_factor

        # Produce the image
        images[:, :, :, 0] *= color[0, :, np.newaxis, np.newaxis]
        images[:, :, :, 1] *= color[1, :, np.newaxis, np.newaxis]
        images[:, :, :, 2] *= color[2, :, np.newaxis, np.newaxis]

        # Apply hexagon mask
        mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
        images *= mask[np.newaxis, :, :, np.newaxis]
        return clip(images)



class ModelF3(Simulator):
    """
    Simulator of the images produced by the light tunnel.

    The derivation of the simulator, including inputs, outputs and
    parameters, is described under Model F1 in Appendix IV.2.2 of the
    paper "Causal chambers as a real-world physical testbed for AI
    Methodology" (2025) by Gamella et al.

    Link for direct access:
    https://arxiv.org/pdf/2404.11341#page=32&zoom=100,57,670
    """

    inputs_names = [
        "red",
        "green",
        "blue",
        "pol_1",
        "pol_2",
    ]
    outputs_names = ["image"]

    def __init__(
        self,
        S,
        w_r,
        w_g,
        w_b,
        exposure,
        Tp,
        Tc,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """
        Initialize the simulator by storing its parameters.
        """
        super(ModelF3, self).__init__()
        # Store the simulator's parameters
        self.S = S
        self.w_r = w_r
        self.w_g = w_g
        self.w_b = w_b
        self.exposure = exposure
        self.Tp = Tp
        self.Tc = Tc
        self.center_x = center_x
        self.center_y = center_y
        self.radius = radius
        self.offset = offset
        self.image_size = image_size

    def parameters(self):
        """
        Return a dictionary with the simulator parameters and their values.
        """
        params = {
            "S": self.S,
            "w_r": self.w_r,
            "w_g": self.w_g,
            "w_b": self.w_b,
            "exposure": self.exposure,
            "Tp": self.Tp,
            "Tc": self.Tc,
            "center_x": self.center_x,
            "center_y": self.center_y,
            "radius": self.radius,
            "offset": self.offset,
            "image_size": self.image_size,
        }
        return params

    def _simulate(
        self,
        red,
        green,
        blue,
        pol_1,
        pol_2,
        S,
        w_r,
        w_g,
        w_b,
        exposure,
        Tp,
        Tc,
        center_x,
        center_y,
        radius,
        offset,
        image_size,
    ):
        """Simulates a synthetic image using model_f3, generating a colored
        hexagon over a black background.

        Parameters
        ----------
        red : np.ndarray
            Brightness of the red LEDs of the light source.
        green : np.ndarray
            Brightness of the green LEDs of the light source.
        blue : np.ndarray
            Brightness of the blue LEDs of the light source.
        pol_1 : np.ndarray
            Angle of the first polarizer.
        pol_2 : np.ndarray
            Angle of the second polarizer.
        S : np.ndarray
            3x3 matrix modeling the sensor response to light of each color.
        w_r : float
            White-balance correction factor for the red channel.
        w_g : float
            White-balance correction factor for the green channel.
        w_b : float
            White-balance correction factor for the blue channel.
        exposure : float
            Exposure parameter affected by camera settings (aperture, shutter speed, ISO).
        Tp : np.ndarray
            Transmission rates for each color when the polarizers are aligned.
        Tc : np.ndarray
            Transmission rates for each color when the polarizers are perpendicular.
        center_x : float
            X-coordinate of the hexagon's center in the image.
        center_y : float
            Y-coordinate of the hexagon's center in the image.
        radius : float
            Radius of the circumference encompassing the hexagon.
        offset : float
            Rotation of the hexagon in degrees.
        image_size : int
            Size of the synthetic image in pixels (i.e., image_size x image_size pixels).

        Returns
        -------
        np.ndarray
            Generated synthetic images with dimensions [n_images,
            image_size, image_size, 3], where n_images is the length of
            the inputs (i.e., red, green, blue, pol_1, pol_2).

        """
        N = len(red)
        images = np.ones((N, image_size, image_size, 3))
        # Transform parameters
        pol_1, pol_2 = np.deg2rad(pol_1), np.deg2rad(pol_2)
        r = np.atleast_1d(red / 255)
        g = np.atleast_1d(green / 255)
        b = np.atleast_1d(blue / 255)

        # Ccalculate color
        malus_factor = (Tp - Tc) * np.cos(pol_1 - pol_2) ** 2 + Tc
        W = np.diag([w_r, w_g, w_b])
        color = exposure * W @ S @ np.array([r, g, b]) * malus_factor

        # Produce the image
        images[:, :, :, 0] *= color[0, :, np.newaxis, np.newaxis]
        images[:, :, :, 1] *= color[1, :, np.newaxis, np.newaxis]
        images[:, :, :, 2] *= color[2, :, np.newaxis, np.newaxis]

        # Apply hexagon mask
        mask = hexagon_mask(center_x, center_y, radius, offset, image_size)
        images *= mask[np.newaxis, :, :, np.newaxis]
        return clip(images)


# --------------------------------------------------------------------
# Auxiliary functions


def clip(images):
    """Clip the pixels of the image so they are always in the range [0,1],
    e.g., 1.2 becomes 1, and -0.1 becomes 0.

    """
    images = np.maximum(images, 0)
    images = np.minimum(images, 1)
    return images


def hexagon_mask(center_x, center_y, radius, offset, image_size):
    """Produce the hexagon mask given its center, radius, (angle) offset
    and the size (in pixels) of the image.

    """
    mask = np.zeros((image_size, image_size))
    image_points = coord_grid(image_size)
    vertices = hexagon_vertices(center_x, center_y, radius, offset) * image_size
    # Compute cross products for a segment and all points
    cross_prods = []
    for i, vertex in enumerate(vertices):
        segment = vertex - vertices[(i + 1) % len(vertices)]
        vertex_to_points = image_points - vertex
        cross = np.cross(vertex_to_points, segment)
        cross_prods.append(cross)
    cross_prods = np.array(cross_prods)
    all_neg = (cross_prods <= 0).all(axis=0)
    all_pos = (cross_prods > 0).all(axis=0)
    mask = np.logical_or(all_neg, all_pos)
    return mask


def hexagon_vertices(center_x, center_y, radius, offset):
    """Given the center, radius and (angle) offset of the hexagon,
    compute the location of its six vertices.

    """
    vertices = []
    for angle in np.arange(0, 2 * np.pi, np.pi / 3):
        x = center_x + radius * np.cos(angle + offset)
        y = center_y + radius * np.sin(angle + offset)
        vertices.append([x, y])
    return np.array(vertices)


def coord_grid(image_size):
    """
    Make a coordinate grid (in pixels) given the image size
    """
    X = np.tile(np.arange(image_size), (image_size, 1))
    Y = X.T[::-1, :]
    grid = np.array([X, Y])
    return np.transpose(grid, (1, 2, 0))



# MIT License

# Copyright (c) 2024 Juan L. Gamella

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


"""
Wildcard module for utility functions.
"""

import numpy as np

# --------------------------------------------------------------------
# Functions to manipulate the ground truth graphs


def edges_to_adjacency(edges, variables):
    """Given edges encoded as a list of tuples, i.e. (from,to), encode as
    an adjacency matrix where `A[from,to] != 0` implies `from -> to`.
    """
    p = len(variables)
    idx = dict(zip(variables, range(p)))
    A = np.zeros((p, p))
    for fro, to in edges:
        if fro in variables and to in variables:
            fro = idx[fro]
            to = idx[to]
            A[fro, to] = 1
    return A


def points_on_circle(n, radius=1):
    """
    Coordinates of n equidistant points on a circle of given radius centered at (0,0).
    """
    theta = np.linspace(0, 2 * np.pi, n + 1)
    x = np.sin(theta)[:-1]
    y = np.cos(theta)[:-1]
    points = list(zip(x, y))
    points.reverse()
    return points


def graph_to_tikz(
    A, points=None, radius=1, labels=None, spaces=4, bend=None, undirected=False
):
    """
    Transform an adjacency matrix A, where `A[i,j] != 0` implies `i -> j` into a latex tikz graph.
    """
    p = len(A)
    if points is None:
        points = points_on_circle(p, radius)
    if labels is None:
        labels = ["$X_{%d}$" % i for i in range(p)]
    indent = " " * spaces
    string = indent + "\\begin{tikzpicture}" + "\n"
    # Add nodes
    for j, (x, y) in enumerate(points):
        string += (
            indent
            + "    \\node[circle, inner sep=0.12em] (%d) at (%0.3f, %0.3f) {%s};\n"
            % (j, x * radius, y * radius, labels[j])
        )
    # Add edges
    string += indent + "    \\begin{scope}[]" + "\n"
    for j, row in enumerate(A):
        for i in np.where(row != 0)[0]:
            if undirected and A[i, j] != 0:
                tip = "-, %s" % undirected
            else:
                tip = "->"
            if bend is None:
                string += indent + "        \\draw[%s] (%d) edge (%d);\n" % (tip, j, i)
            else:
                jj, ii = 0, (i - j) % p
                direction = "left" if (ii - jj) < p / 2 else "right"
                amount = abs(bend * (1 - (ii - jj) / p * 2))
                bend_str = "bend %s=%d" % (direction, amount)
                string += indent + "        \\draw[%s, %s] (%d) edge (%d);\n" % (
                    tip,
                    bend_str,
                    j,
                    i,
                )
    string += indent + "    \\end{scope}" + "\n"
    string += indent + "\\end{tikzpicture}" + "\n"
    return string



from .wind_tunnel_models import (
    model_a1,
    model_b1,
    model_a2,
    model_c1,
    model_c2,
    model_c3,
)
from .wind_tunnel_simulators import simulator_a1_c2, simulator_a1_c3, simulator_a2_c3
from .image_capture import model_f1, model_f2, model_f3
from .light_tunnel_models import model_e1



#!/usr/bin/env python
# coding: utf-8

# # **Tutorial**: Using the simulators `lt.ModelF1`, `lt.ModelF2` and `lt.ModelF3`
# 
# [<<< Back to the simulator index](https://github.com/juangamella/causal-chamber-package/tree/main/causalchamber/simulators)
# 
# Other links:
# - [Derivation of the simulators](https://arxiv.org/pdf/2404.11341#page=32&zoom=100,57,670)
# - [Simulators source code](https://github.com/juangamella/causal-chamber-package/blob/main/causalchamber/simulators/lt/image/models_f.py)
# 
# 
# If you use the simulators in your work, please consider citing:
# 
# ```
# ﻿@article{gamella2025chamber,
#   author={Gamella, Juan L. and Peters, Jonas and B{\"u}hlmann, Peter},
#   title={Causal chambers as a real-world physical testbed for {AI} methodology},
#   journal={Nature Machine Intelligence},
#   doi={10.1038/s42256-024-00964-x},
#   year={2025},
# }
# ```

# In[1]:


import sys
sys.path = ['../../..'] + sys.path

# %load_ext autoreload
# %autoreload 2


# ## Dependencies

# Besides the `causalchamber` package, the notebook also uses `matplotlib`, `numpy` and `pandas` (specified in `tutorials/requirements.txt`).

# In[2]:


import causalchamber
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from matplotlib import gridspec


# ## Initializing the simulators
# 
# We take the simulator parameters from calibration experiments and the technical specifications of the camera and polarizers. See [this](https://github.com/juangamella/causal-chamber-paper/blob/main/case_studies/mechanistic_models.ipynb) notebook for the complete process.

# In[3]:


# All parameters
S = np.array(
    [
        [0.2995802, 0.06920322, 0.028159],
        [0.04017132, 0.9209918, 0.31265],
        [0.00217392, 0.2071558, 0.75763],
    ]
)
w_r = 2.65625
w_g = 1.0
w_b = 1.77344
Tp = np.array([[0.29, 0.35, 0.33]]).T
Tc = np.array([[0.02, 0.08, 0.18]]).T
center_x = 0.5
center_y = 0.5
radius = 0.22
offset = 0
image_size = 500


# Initialize simulators

# In[4]:


import causalchamber.simulators.lt as lt


# In[5]:


# Model F1
sim_f1 = lt.ModelF1(center_x=center_x,
                    center_y=center_y,
                    radius=radius,
                    offset=offset,
                    image_size=image_size)

# Model F2
sim_f2 = lt.ModelF2(S=S,
                    w_r=w_r,
                    w_g=w_g,
                    w_b=w_b,
                    exposure=0.8,
                    center_x=center_x,
                    center_y=center_y,
                    radius=radius,
                    offset=offset,
                    image_size=image_size)

# Model F3
sim_f3 = lt.ModelF3(S=S,
                    w_r=w_r,
                    w_g=w_g,
                    w_b=w_b,
                    Tp=Tp,
                    Tc=Tc,
                    exposure=2.5,
                    center_x=center_x,
                    center_y=center_y,
                    radius=radius,
                    offset=offset,
                    image_size=image_size)


# ## Running the simulators

# To simulate, call the function `.simulate_from_inputs` with a dataframe containing the simulator inputs as columns. There can be additional columns, which the simulator will ignore.
# 
# The names of the necessary columns can be obtained as:

# In[6]:


sim_f1.inputs_names


# ### Example 1: Random inputs
# 
# For example, let's generate some images from random inputs:

# In[7]:


rng = np.random.default_rng(7); N = 5

inputs = pd.DataFrame({'red': rng.uniform(0,255,size=N),
                       'green': rng.uniform(0,255,size=N),
                       'blue': rng.uniform(0,255,size=N),
                       'pol_1': rng.uniform(-180,180,size=N),
                       'pol_2': rng.uniform(-180,180,size=N)})

outputs_f1 = sim_f1.simulate_from_inputs(inputs)
outputs_f2 = sim_f2.simulate_from_inputs(inputs)
outputs_f3 = sim_f3.simulate_from_inputs(inputs)


# Plot the results

# In[8]:


plt.figure(figsize=(N*2,6))
for i,(output_f1, output_f2, output_f3) in enumerate(zip(outputs_f1, outputs_f2, outputs_f3)):
    plt.subplot(3,N,i+1)
    plt.imshow(output_f1)
    plt.title('red=%d\ngreen=%d\nblue=%d\npol_1=%0.1f\npol_2=%0.1f' % tuple(inputs.iloc[i][sim_f1.inputs_names]))
    plt.axis("off")
    plt.subplot(3,N,i+1+N)
    plt.imshow(output_f2)
    plt.axis("off")
    plt.subplot(3,N,i+1+2*N)
    plt.imshow(output_f3)
    plt.axis("off")


# ### Example 2: Compare to real images

# We download a dataset of real images using the functionality provided in the `causalchamber` package. The downloaded dataset is stored in the path given by `root=` below.
# 
# For our experiment we will use the `lt_camera_test_v1` dataset, which is described [here](https://github.com/juangamella/causal-chamber/tree/main/datasets/lt_camera_test_v1).

# In[9]:


# Download the dataset
dataset = causalchamber.datasets.Dataset('lt_camera_test_v1', root='./', download=True)

# To highlight the difference between simulators, we select some specific images from the experiments "palette" and "polarizer_effects_bright"
images_palette = dataset.get_experiment("palette").as_image_array('500')[0:15]
images_polarizers = dataset.get_experiment("polarizer_effect_bright").as_image_array('500')
df_palette = dataset.get_experiment("palette").as_pandas_dataframe().iloc[0:15]
df_polarizers = dataset.get_experiment("polarizer_effect_bright").as_pandas_dataframe()

# Final images / observations
idx = [11,12,-1] 
images = np.concatenate([images_palette, images_polarizers], axis=0)[idx]
observations = pd.concat([df_palette, df_polarizers], ignore_index=True).iloc[idx]


# We can directly use the `observations` dataframe as input to the simulators, since it contains the necessary columns. The additional columns are ignored by the simulator.

# In[10]:


outputs_f1 = sim_f1.simulate_from_inputs(observations)
outputs_f2 = sim_f2.simulate_from_inputs(observations)
outputs_f3 = sim_f3.simulate_from_inputs(observations)


# Plot the results

# In[11]:


plt.figure(figsize=(4,len(images)*1), dpi=200)
gs = gridspec.GridSpec(len(images), 4, wspace=0, hspace=0)#, height_ratios = [.22, .26, .26, .26])

for i,(real_image,output_f1,output_f2,output_f3) in enumerate(zip(images, outputs_f1, outputs_f2, outputs_f3)):
    
    plt.subplot(gs[i,0])
    ax = plt.gca()
    plt.imshow(real_image)
    plt.yticks([])
    plt.xticks([])
    
    plt.subplot(gs[i,1])
    ax = plt.gca()
    plt.imshow(output_f1)
    plt.yticks([])
    plt.xticks([])
    
    plt.subplot(gs[i,2])
    ax = plt.gca()
    plt.imshow(output_f2)
    plt.yticks([])
    plt.xticks([])
    
    plt.subplot(gs[i,3])
    ax = plt.gca()
    plt.imshow(output_f3)
    plt.yticks([])
    plt.xticks([])


# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# # **Tutorial**: Using the simulator `lt.DecoderSimple`
# 
# [<<< Back to the simulator index](https://github.com/juangamella/causal-chamber-package/tree/main/causalchamber/simulators)
# 
# Other links:
# - [Derivation of the simulators](TODO) -> TODO
# - [Simulators source code](https://github.com/juangamella/causal-chamber-package/blob/main/causalchamber/simulators/lt/image/decoder.py)
# 
# If you use the simulator in your work, please consider citing:
# 
# ```
# TODO
# ```

# In[1]:


import sys
sys.path = ['../../../'] + sys.path

# %load_ext autoreload
# %autoreload 2


# ## Dependencies

# This simulator requires PyTorch as a dependency. This includes the Python packages `torch` and `torchvision`, specified in the `tutorials/requirements.txt` file and as additional dependencies to the `causalchamber` package (option `[torch]`).
# 
# The notebook also uses `matplotlib`, `numpy` and `pandas` (also specified in `tutorials/requirements.txt`).

# In[2]:


import causalchamber
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


# ## Initializing the simulator
# 
# First, import and instantiate the simulator. This will download the pre-trained model into the path given by `root=`.

# In[3]:


import causalchamber.simulators.lt as lt
simulator = lt.DecoderSimple(root='./') # store pre-trained model in the current directory


# ## Running the simulator

# To simulate, call the function `.simulate_from_inputs` with a dataframe containing the simulator inputs as columns. There can be additional columns, which the simulator will ignore.
# 
# The names of the necessary columns can be obtained as:

# In[4]:


simulator.inputs_names


# ### Example 1: Random inputs
# 
# For example, let's generate some images from random inputs:

# In[ ]:


rng = np.random.default_rng(7); N = 5

inputs = pd.DataFrame({'red': rng.uniform(0,255,size=N),
                       'green': rng.uniform(0,255,size=N),
                       'blue': rng.uniform(0,255,size=N),
                       'pol_1': rng.uniform(-180,180,size=N),
                       'pol_2': rng.uniform(-180,180,size=N)})

outputs = simulator.simulate_from_inputs(inputs)


# Plot the results

# In[6]:


plt.figure(figsize=(N*2,2))
for i,output in enumerate(outputs):
    plt.subplot(1,N,i+1)
    plt.imshow(output)
    plt.title('red=%d\ngreen=%d\nblue=%d\npol_1=%0.1f\npol_2=%0.1f' % tuple(inputs.iloc[i][simulator.inputs_names]))
    plt.axis("off")


# ### Example 2: Compare to real images

# We download a dataset of real images using the functionality provided in the `causalchamber` package. The downloaded dataset is stored in the path given by `root=` below.
# 
# For our experiment we will use the `lt_crl_benchmark_v1` dataset, which is described [here](TODO).

# In[7]:


# Download the dataset
dataset = causalchamber.datasets.Dataset('lt_crl_benchmark_v1', root='./', download=True)

# Load the data (inputs, measurements and images) from one of the experiments
experiment = dataset.get_experiment(name='uniform')
observations = experiment.as_pandas_dataframe()
images = experiment.as_image_array(size='64')


# We can directly use the `observations` dataframe as input to the simulator, since it contains the necessary columns. The additional columns are ignored by the simulator.

# In[8]:


inputs = observations.iloc[[56, 78, 85, 90, 54, 88]] # Select some images to get a nice variation of colors
outputs = simulator.simulate_from_inputs(inputs)


# Plot the results

# In[9]:


K = len(outputs)
plt.figure(figsize=(K*2,4))
for i,output in enumerate(outputs):
    plt.subplot(2,K,i+1)
    plt.imshow(images[inputs.index[i]])
    plt.title("Real")
    plt.axis("off")
    plt.subplot(2,K,K+i+1)
    plt.imshow(output)
    plt.title("Simulated")
    plt.axis("off")


# In[ ]:




